from __future__ import division

import logging


import matplotlib as mpl
import SimpleITK as sitk
import numpy as np


def convert_discrete_colormap(data, cm_name='custom', color_names=None):
    '''Generates a matplotlib continuous colormap on [0, 1] from a discrete 
    colormap at N evenly spaced points.

    Parameters
    ----------
    data : list of list
        Sublists are [r, g, b]. 
    
    Returns
    -------
    matplotlib.colors.LinearSegmentedColormap
        Gamma is 1. Output space is 3 X [0, 1]
    
    '''

    if color_names is None:
        color_names = ['red', 'green', 'blue']

    data = np.array(data)
    npoints = data.shape[0]

    domain = np.linspace(0, 1.0, npoints)
    color_arrays = {}

    for col, name in enumerate(color_names):
        color_array = np.zeros((npoints, 3))
        
        color_array[:, 0] = domain
        color_array[:, 1] = minmax_norm(data[:, col])
        color_array[:, 2] = color_array[:, 1]

        color_arrays[name] = color_array

    return mpl.colors.LinearSegmentedColormap(cm_name, color_arrays, npoints, gamma=1.0)



def minmax_norm(data):

    rng = np.amax(data) - np.amin(data)
    if rng == 0:
        return data 

    return (data - np.amin(data)) / rng



def sitk_safe_ln(data, minimum=10**-10):

    logging.info('thresholding below at {0}'.format(minimum))
    minimum = float(minimum)
    data = sitk.Threshold(data, minimum, np.inf, minimum)

    logging.info('taking natural log')
    return sitk.Log(data)


def normalize_intensity(data, in_min, in_max, out_min=0.0, out_max=0.0):

    logging.info('setting input range: [{0:2.3f}, {1:2.3f}]'.format(in_min, in_max))
    data = sitk.ShiftScale(data, -in_min, 1.0 / (in_max - in_min))
    data = sitk.Threshold(data, 0.0, np.inf, 0.0)
    data = sitk.Threshold(data, 0.0, 1.0, 1.0)

    logging.info('setting output range: [{0:2.3f}, {1:2.3f}]'.format(out_min, out_max))
    data = sitk.ShiftScale(data, 0.0, out_max - out_min) # want to scale first
    return sitk.ShiftScale(data, out_min, 1) # then shift


def blend(image_stack, weight_stack):
    '''

    Parameters
    ----------
    image_stack :: list of np.ndarray
        The images to be blended. Shapes cannot differ
    weight_stack :: list of np.ndarray
        The weight of each image at each pixel. Will be normalized.

    '''

    image_stack = np.array(image_stack)
    weight_stack = np.array(weight_stack)

    weight_stack = weight_stack - np.amin(weight_stack, axis=0) / (np.amax(weight_stack, axis=0) - np.amin(weight_stack, axis=0))
    weight_stack[np.isnan(weight_stack)] = 0.5
    weight_stack[np.isinf(weight_stack)] = 0.5

    return np.multiply(image_stack, weight_stack).sum(axis=0)



import logging
import sys
import os
import h5py
import subprocess
import shutil
import numpy as np
import traceback
import nwb.nwb as nwb
import nwb.nwbco as nwbco
import resource_file
from collections import defaultdict

from allensdk.core.nwb_data_set import NwbDataSet
import allensdk.ephys.extract_cell_features as extract_cell_features

from allensdk.internal.core.lims_pipeline_module import PipelineModule

# changes
"""
added to main block:
    "nwb_file": "igor_converted_256189.05.01.nwb",
    "publish_nwb": "to_publish.nwb",
    "metadata_file": "nwb_metadata.yml",


"""

local_dir = os.path.dirname(os.path.realpath(__file__))
if not local_dir.endswith('/'):
    local_dir += '/'


ELECTRODE_NAME = "Electrode 1"
ELECTRODE_PATH = "/general/intracellular_ephys/" + ELECTRODE_NAME

PIPELINE_NAME = "IVSCC"
PIPELINE_VERSION = "1.0"

def copy_val(old_ts, new_ts, name):
    if name in old_ts:
        val = old_ts[name].value
        attrs = {}
        for x in old_ts[name].attrs:
            if x == "neurodata_type":
                continue
            if x == "unit":
                continue
            attrs[x] = old_ts[name].attrs[x]
        new_ts.set_value(name, val, **attrs)

def copy_timeseries(timeseries, old_file, new_file, folder, metadata):
    try:
        name = ""
        for name in timeseries:
            old_ts = old_file[folder][name]
            family = old_ts.attrs["ancestry"]
            if family[-1] == "VoltageClampSeries":
                family = family[-1]
                category = "acquisition"
            elif family[-1] == "CurrentClampSeries":
                family = family[-1]
                category = "acquisition"
            elif family[-1] == "VoltageClampStimulusSeries":
                family = family[-1]
                category = "stimulus"
            elif family[-1] == "CurrentClampStimulusSeries":
                family = family[-1]
                category = "stimulus"
            else:
                raise Exception("Time series '%s' is of unknown type" % name)
            new_ts = new_file.create_timeseries(family, name, category)
            # copy data
            num_samples = old_ts["num_samples"].value
            data = old_ts["data"].value
            conversion = old_ts["data"].attrs["conversion"]
            resolution = old_ts["data"].attrs["resolution"]
            unit = old_ts["data"].attrs["unit"]
            new_ts.set_data(data, conversion=conversion, resolution=resolution, unit=unit)

            start_time = old_ts["starting_time"].value
            sampling_rate = old_ts["starting_time"].attrs["rate"]
            new_ts.set_time_by_rate(start_time, sampling_rate)
            new_ts.set_value("num_samples", num_samples)

            description = old_ts.attrs["description"]
            try:
                comments = old_ts.attrs["comments"]
            except:
                comments = old_ts.attrs["comment"]
            source = old_ts.attrs["source"]
            new_ts.set_value("comments", comments)
            new_ts.set_value("description", description)
            new_ts.set_value("source", source)
            
            copy_val(old_ts, new_ts, "electrode_name")
            copy_val(old_ts, new_ts, "capacitance_fast")
            copy_val(old_ts, new_ts, "capacitance_slow")
            copy_val(old_ts, new_ts, "resistance_comp_bandwidth")
            copy_val(old_ts, new_ts, "resistance_comp_correction")
            copy_val(old_ts, new_ts, "resistance_comp_prediction")
            copy_val(old_ts, new_ts, "whole_cell_capaictance_comp")
            copy_val(old_ts, new_ts, "whole_cell_series_resistance_comp")
            copy_val(old_ts, new_ts, "bias_current")
            copy_val(old_ts, new_ts, "bridge_balance")
            copy_val(old_ts, new_ts, "capacitance_compensation")
            copy_val(old_ts, new_ts, "stimulus_description")
            # 
            new_ts.finalize()
    except:
        print("** Error copying timeseries data **")
        print("** Timeseries: " + str(name))
        print("** Folder: " + folder)
        print("-----------------------------------")
        raise

def copy_epochs(timeseries, old_file, new_file, folder):
    try:
        for name in timeseries:
            anc = old_file["acquisition/timeseries/"+name].attrs["ancestry"]
            if anc[-1] == "VoltageClampSeries":
                continue
            num = int(name.split('_')[-1])
            # experiment block
            epname = "Experiment_%d" % num
            ep = old_file["epochs/%s" % epname]
            start = ep["start_time"].value
            stop = ep["stop_time"].value
            desc = ep["description"].value
            ep = new_file.create_epoch(epname, start, stop)
            ep.set_value("description", desc)
            ep.add_timeseries("stimulus", "/stimulus/presentation/Sweep_%d" % num)
            ep.add_timeseries("response", "/acquisition/timeseries/Sweep_%d" % num)
            ep.finalize()
            # test-pulse block
            epname = "TestPulse_%d" % num
            ep = old_file["epochs/%s" % epname]
            start = ep["start_time"].value
            stop = ep["stop_time"].value
            desc = ep["description"].value
            ep = new_file.create_epoch(epname, start, stop)
            ep.set_value("description", desc)
            ep.add_timeseries("stimulus", "/stimulus/presentation/Sweep_%d" % num)
            ep.add_timeseries("response", "/acquisition/timeseries/Sweep_%d" % num)
            ep.finalize()
            # sweep block
            epname = name
            ep = old_file["epochs/%s" % epname]
            start = ep["start_time"].value
            stop = ep["stop_time"].value
            desc = ep["description"].value
            ep = new_file.create_epoch(epname, start, stop)
            ep.set_value("description", desc)
            ep.add_timeseries("stimulus", "/stimulus/presentation/Sweep_%d" % num)
            ep.add_timeseries("response", "/acquisition/timeseries/Sweep_%d" % num)
            ep.finalize()

    except:
        print("** Error copying epoch data **")
        print("------------------------------")
        raise


def copy_file(infile, outfile, passing_sweeps, rsrc, metadata):
    print("Opening '%s'" % infile)
    old = h5py.File(infile, 'r')
    # top-level data
    try:
        vargs = {}
        vargs["identifier"] = old["identifier"].value[0] + ".edit"
        vargs["start_time"] = old["session_start_time"].value[0]
        vargs["description"] = old["session_description"].value[0]
        vargs["overwrite"] = True
        vargs["filename"] = outfile
        vargs["auto_compress"] = True
    except:
        print("** Error extracting top-level metadata from input file **")
        print("---------------------------------------------------------")
        raise

    print("Creating '%s'" % outfile)
    try:
        out = nwb.NWB(**vargs)
    except:
        print("** Error creating output file '%s' **" % outfile)
        print("---------------------------------------------------------")
        raise

    # make list of time series
    timeseries = []
    try:
        acq = old["acquisition/timeseries"]
        for ts in acq:
            timeseries.append(ts)
    except:
        print("** Error extracting timeseries list **")
        print("--------------------------------------")
        raise

    ts_list = []
    # TODO remove items from list that are not to be copied
    for num in passing_sweeps:
        swp = "Sweep_%d" % num
        for name in timeseries:
            if name == swp:
                ts_list.append(swp)
                break
    timeseries = ts_list

    # copy acquisition time series from source to destination files
    copy_timeseries(timeseries, old, out, "acquisition/timeseries", metadata)
    copy_timeseries(timeseries, old, out, "stimulus/presentation", metadata)

    copy_epochs(timeseries, old, out, "stimulus/presentation")

    write_metadata(out, rsrc, metadata)
        
    out.close()

def organize_metadata(ephys_roi_result):
    metadata = { 'sweeps': {} }
    
    cell_specimen = ephys_roi_result['specimens'][0]
    slice_specimen = ephys_roi_result['specimen']

    metadata['donor_id'] = cell_specimen['donor_id']
    metadata['specimen_name'] = cell_specimen['name']
    metadata['specimen_id'] = cell_specimen['id']
    try:
        metadata['species'] = slice_specimen['donor']['organism']["name"]
    except Exception as e:
        logging.error("Unable to read organism name from input.json file")
        raise

    # structure
    try:
        structure = cell_specimen['structure']
    except Exception as e:
        logging.error("Cell has no structure association.")
        raise
    
    soma_location = {}

    db_cell_soma_location = cell_specimen['cell_soma_locations'][0]
    soma_location = {}
    try:
        soma_location['cell_soma_location_x'] = 1e-9 * db_cell_soma_location['x']
        soma_location['cell_soma_location_y'] = 1e-9 * db_cell_soma_location['y']
        soma_location['cell_soma_location_z'] = 1e-9 * db_cell_soma_location['z']
        nd = db_cell_soma_location['normalized_depth']
        if nd is not None:
            soma_location['cell_soma_location_normalized_depth'] = nd 
    except Exception as e:
        logging.error(e.message)
        raise

    structure_info = {}
    try:
        structure_info['structure_id'] = structure['id']
        structure_info['structure_name'] =  structure['name']
        structure_info['structure_acronym'] = structure['acronym']
    except Exception as e:
        logging.error("Structure information is missing from input.json")
        raise 

    structure_info.update(soma_location)
    metadata['location'] = structure_info


    tags = cell_specimen["specimen_tags"]

    dend_trunc = None
    dend_type = None
    for i in range(len(tags)):
        name = tags[i]["name"]
        toks = name.split(" - ")
        if len(toks) != 2:
            continue
        if name.startswith("apical"):
            dend_trunc = toks[1]
        elif name.startswith("dendrite type"):
            dend_type = toks[1]

    if dend_trunc is None:
        raise Exception("Cell has no dendrite truncation tag.")

    if dend_type is None:
        raise Exception("Cell has no dendrite type tag.")

    metadata['dendrite_type'] = dend_type
    metadata['dendrite_trunc'] = dend_trunc
    
    metadata['ephys_roi_result_id'] = ephys_roi_result['id']
    metadata['seal_gohm'] = ephys_roi_result['seal_gohm']
    metadata['initial_access_resistance_mohm'] = ephys_roi_result['initial_access_resistance_mohm']

    slice_specimen = ephys_roi_result['specimen']
    donor = slice_specimen['donor']

    # gender
    try:
        metadata['gender'] = donor['gender']['name']
    except Exception as e:
        logging.error("Donor requires gender association.")
        raise

    # age
    try:
        age = donor['age']
    except Exception as e:
        logging.error("Donor requires age association.")
        raise

    metadata['age'] = {
        'date_of_birth': donor['date_of_birth'],
        'name': age['name']
        }


    # cre line and genotype are mouse-only
    if metadata['species'] == 'Mus musculus':
        genotypes = donor['genotypes']

        try:
            reporter_genotype = next( g for g in genotypes if g['genotype_type_id'] == 177835595 )
            metadata['cre_line'] = reporter_genotype['name']
        except Exception, e:
            logging.error("Could not find reporter genotype for mouse cell")
            raise

        metadata['genotype'] = {
            'description': [ g['description'] for g in genotypes ],
            'type': [ g['name'] for g in genotypes ]
            }
    else:
        logging.info("non-mouse cells do not have cre line or genotypes")

    # subject
    metadata['subject'] = {
        'subject_id': cell_specimen['donor_id'],
        'comments': 'subject_id value here corresponds to Allen Institute cell specimen "donor_id"'
        }

    # sweeps
    sweeps = cell_specimen['ephys_sweeps']
    for sweep in sweeps:
        if "invalid" in sweep and sweep["invalid"]:
            continue
        wfs = sweep['workflow_state']
        if wfs not in [ 'manual_passed', 'auto_passed' ]:
            logging.debug("skipping sweep %d, not passed" % sweep['sweep_number'])
            continue
            
        stimulus = sweep['ephys_stimulus']
        stimulus_type = stimulus['ephys_stimulus_type']


        metadata['sweeps'][sweep['sweep_number']] = {
            'stimulus_name': stimulus['description'],
            'stimulus_interval': sweep['stimulus_interval'],
            'stimulus_amplitude': sweep['stimulus_amplitude'],
            'stimulus_type_name': stimulus_type['name']
            }

    # IT-12498 add additional metadata to NWB file
    url = "http://help.brain-map.org/display/celltypes/Documentation"
    metadata["data_collection"] = "please see " + url
    metadata["protocol"] = "please see " + url
    metadata["pharmacology"] = "please see " + url
    metadata["citation_policy"] = "please see " + url
    metadata["institution"] = "Allen Institute for Brain Science"
    metadata["generated_by"] = ["pipeline", PIPELINE_NAME, "version", PIPELINE_VERSION]

    return metadata


def write_metadata(nwb_file, resources, metadata):
    nwb_file.set_metadata(nwbco.SEX, metadata['gender'])
    if 'cre_line' in metadata:
        nwb_file.set_metadata("aibs_cre_line", metadata['cre_line'])

    if 'genotype' in metadata:
        genotype = metadata['genotype']
        genotype_name = '; '.join(genotype['type'])
        nwb_file.set_metadata(nwbco.GENOTYPE, genotype_name, **genotype)

    nwb_file.set_metadata('generated_by', metadata['generated_by'])

    subject = metadata['subject']
    nwb_file.set_metadata(nwbco.SUBJECT, resources.get("subject"), **subject)

    age = metadata['age']
    nwb_file.set_metadata(nwbco.AGE, age['name'], **age)

    trode = ELECTRODE_NAME
    nwb_file.set_metadata(nwbco.INTRA_ELECTRODE_DESCRIPTION(trode), resources.get("electrode_description"))
    nwb_file.set_metadata(nwbco.INTRA_ELECTRODE_FILTERING(trode), resources.get("electrode_filtering"))
    nwb_file.set_metadata(nwbco.INTRA_ELECTRODE_DEVICE(trode), resources.get("electrode_device"))

    location = metadata['location']
    nwb_file.set_metadata(nwbco.INTRA_ELECTRODE_LOCATION(trode), location['structure_name'], **location)

    nwb_file.set_metadata(nwbco.INTRA_ELECTRODE_RESISTANCE(trode), resources.get("electrode_resistance"))
    nwb_file.set_metadata(nwbco.INTRA_ELECTRODE_SLICE(trode), resources.get("electrode_slice"))

    seal_gohm = str(metadata['seal_gohm'])
    nwb_file.set_metadata(nwbco.INTRA_ELECTRODE_SEAL(trode), seal_gohm + " GOhm")

    acc = str(metadata["initial_access_resistance_mohm"])
    nwb_file.set_metadata(nwbco.INTRA_ELECTRODE_INIT_ACCESS_RESISTANCE(trode), acc + " MOhm")

    session = { 'comments': 'session_id value corresponds to ephys_result_id' }
    nwb_file.set_metadata(nwbco.SESSION_ID, str(metadata['ephys_roi_result_id']), **session)

    nwb_file.set_metadata("aibs_specimen_name", metadata['specimen_name'])
    nwb_file.set_metadata("aibs_specimen_id", str(metadata['specimen_id']))

    nwb_file.set_metadata("aibs_dendrite_type", metadata['dendrite_type'])
    nwb_file.set_metadata("aibs_dendrite_trunc", metadata['dendrite_trunc'])    

    # IT-12498 add additional metadata to NWB file
    nwb_file.set_metadata(nwbco.DATA_COLLECTION, metadata["data_collection"])
    nwb_file.set_metadata(nwbco.INSTITUTION, metadata["institution"])
    nwb_file.set_metadata(nwbco.PROTOCOL, metadata["protocol"])
    nwb_file.set_metadata(nwbco.PHARMACOLOGY, metadata["pharmacology"])
    nwb_file.set_metadata("citation_policy", metadata["citation_policy"])
    nwb_file.set_metadata(nwbco.SPECIES, metadata['species'])



def main(jin):
    infile = jin[0]["nwb_file"]
    outfile = jin[0]["publish_nwb"]
    tmpfile = outfile + ".working"
    metafile = local_dir + jin[0]["metadata_file"]
    # load metadata stored in YML file
    metadata_desc_file = os.path.join(os.path.dirname(__file__), metafile)
    rsrc = resource_file.ResourceFile()
    rsrc.load(metadata_desc_file)

    #
    metadata = organize_metadata(jin[0])

    # TODO dig deeper here
    # only fetching metadata for passing sweeps
    passing_sweeps = metadata['sweeps'].keys()

    copy_file(infile, outfile, passing_sweeps, rsrc, metadata)

#    try:
#        shutil.copyfile(infile, tmpfile)
#    except:
#        print("Unable to copy '%s' to %s" % (infile, tmpfile))
#        print("----------------------------")
#        raise

#    # open NWB file so the modification date is updated
#    # add metadata then close file and do remaining manipulations using 
#    #   HDF5 library (except DF's legacy code that interfaces w/ nwb file
#    #   using nwb library)
#    args = {}
#    args["filename"] = tmpfile
#    args["modify"] = True
#    try:
#        nwb_file = nwb.NWB(**args)
#    except:
#        print("Error opening NWB file '%s'" % args["filename"])
#        raise
#    write_metadata(nwb_file, rsrc, metadata)
#    nwb_file.close()



    # open publish file directlya using HDF5 library
    # 1) remove hdf5 groups corresponding to failed sweeps
    # 2) add sweep-specific metadata data to file to match original publish
    #   format. this includes (acquisition and stimulus):
    #     aibs_stimulus_amplitude_pa
    #     aibs_stimulus_interval
    #     aibs_stimulus_name
    #     initial_access_resistance
    #     seal
    hdf = h5py.File(outfile, "r+")
#    ################################
#    # delete epochs, stim, recordings for non-passed sweeps
#    epochs = hdf["epochs/"]
#    for grp in epochs:
#        try:
#            num = int(str(grp).split('_')[-1])
#        except:
#            continue
#        if num not in passing_sweeps:
#            del epochs[str(grp)]
#    stim = hdf["stimulus/presentation"]
#    for grp in stim:
#        try:
#            num = int(str(grp).split('_')[-1])
#        except:
#            continue
#        if num not in passing_sweeps:
#            del stim[str(grp)]
#    acq = hdf["acquisition/timeseries"]
#    for grp in acq:
#        try:
#            num = int(str(grp).split('_')[-1])
#        except:
#            continue
#        if num not in passing_sweeps:
#            del acq[str(grp)]
    ################################
    # add data
    acq = hdf["acquisition/timeseries"]
    stim = hdf["stimulus/presentation"]
    sweeps = jin[0]["specimens"][0]["ephys_sweeps"]
    for grp in acq:
        try:
            num = int(str(grp).split('_')[-1])
        except:
            continue
        try:
            for sweep in sweeps:
                if sweep["sweep_number"] == num:
                    break
            if sweep["sweep_number"] != num:
                print sweep
                print num
                raise Exception("WTF")
            # stim amplitude
            amp = sweep["stimulus_amplitude"]
            if amp is None:
                amp = float('nan')
            else:
                amp = float(amp)
            ds = acq["Sweep_%d" % num].create_dataset("aibs_stimulus_amplitude_pa", data=amp)
            ds.attrs["neurodata_type"] = "Custom"
            ds = stim["Sweep_%d" % num].create_dataset("aibs_stimulus_amplitude_pa", data=amp)
            ds.attrs["neurodata_type"] = "Custom"
            # stim interval
            interval = sweep["stimulus_interval"]
            if interval is None:
                interval = float('nan')
            else:
                interval = float(interval)
            ds = acq["Sweep_%d" % num].create_dataset("aibs_stimulus_interval", data=interval)
            ds.attrs["neurodata_type"] = "Custom"
            ds = stim["Sweep_%d" % num].create_dataset("aibs_stimulus_interval", data=interval)
            ds.attrs["neurodata_type"] = "Custom"
            # stim name
            name = sweep["ephys_stimulus"]["ephys_stimulus_type"]["name"]
            ds = acq["Sweep_%d" % num].create_dataset("aibs_stimulus_name", data=name)
            ds.attrs["neurodata_type"] = "Custom"
            ds = stim["Sweep_%d" % num].create_dataset("aibs_stimulus_name", data=name)
            ds.attrs["neurodata_type"] = "Custom"
            # seal
            seal = jin[0]["seal_gohm"]
            if seal is None:
                seal = float('nan')
            else:
                seal = float(seal)
            ds = acq["Sweep_%d" % num].create_dataset("seal", data=seal)
            ds.attrs["neurodata_type"] = "Custom"
            ds = stim["Sweep_%d" % num].create_dataset("seal", data=seal)
            ds.attrs["neurodata_type"] = "Custom"
            # initial access resistance
            res = jin[0]["initial_access_resistance_mohm"]
            if res is None:
                res = float('nan')
            else:
                res = float(res)
            ds = acq["Sweep_%d" % num].create_dataset("initial_access_resistance", data=res)
            ds.attrs["neurodata_type"] = "Custom"
            ds = stim["Sweep_%d" % num].create_dataset("initial_access_resistance", data=res)
            ds.attrs["neurodata_type"] = "Custom"
            # 
#            # recycle code from old publish module for custom sweep metadata
#            if num in metadata['sweeps']:
#                sweep_md = metadata['sweeps'][num]
#                stimulus_interval = sweep_md['stimulus_interval']
#                if stimulus_interval is None:
#                    stimulus_interval = float('nan')
#                ds = acq["Sweep_%d" % num].create_dataset("aibs_stimulus_interval", data=stimulus_interval)
#                ds.attrs["neurodata_type"] = "Custom"
#                #
#                ds = acq["Sweep_%d" % num].create_dataset("aibs_stimulus_name", data=sweep_md['stimulus_type_name'])
#                ds.attrs["neurodata_type"] = "Custom"
#                #
#                ds = acq["Sweep_%d" % num].create_dataset("aibs_stimulus_amplitude_%s" % stim_units, sweep_md['stimulus_amplitude'])
#                ds.attrs["neurodata_type"] = "Custom"
#                #
#                ds = acq["Sweep_%d" % num].create_dataset("seal", sweep_md['seal_gohm'])
#                ds.attrs["neurodata_type"] = "Custom"
        except:
            print("json parse error for sweep %d" % num)
            raise
    # all done
    hdf.close()


    # TODO describe what's happening here
    sweeps_by_type = defaultdict(list)
    for sweep_number, sweep_data in metadata['sweeps'].iteritems():
        sweeps_by_type[sweep_data['stimulus_type_name']].append(sweep_number)

    sweep_features = extract_cell_features.extract_sweep_features(NwbDataSet(outfile), sweeps_by_type)

    # TODO describe what's happening here
    for sweep_num in passing_sweeps:
        try:
            spikes = sweep_features[sweep_num]['spikes']
            spike_times = [ s['threshold_t'] for s in spikes ]
            NwbDataSet(outfile).set_spike_times(sweep_num, spike_times)
        except Exception, e:
            logging.info("sweep %d has no sweep features. %s" % (sweep_num, e.message) )
#    try:
#        # remove spike times for non-passing sweeps
#        spk = hdf["analysis/spike_times"]
#        for grp in spk:
#            try:
#                num = int(str(grp).split('_')[-1])
#            except:
#                continue
#            if num not in passing_sweeps:
#                del spk[str(grp)]
#    except:
#        

#    # rescaling the contents of the data arrays causes the file to grow
#    # execute hdf5-repack to get it back to its original size
#    try:
#        print("Repacking hdf5 file with compression")
#        process = subprocess.Popen(["h5repack", "-f", "GZIP=4", tmpfile, outfile], stdout=subprocess.PIPE)
#        process.wait()
#    except:
#        print("Unable to run h5repack on temporary nwb file")
#        print("--------------------------------------------")
#        raise

#    try:
#        print("Removing temporary file")
#        os.remove(tmpfile)
#    except:
#        print("Unable to delete temporary file ('%s')" % tmpfile)
#        raise

    empty = {}
    return empty


if __name__ == "__main__": 
    # read module input. PipelineModule object automatically parses the 
    #   command line to pull out input.json and output.json file names
    module = PipelineModule()
    jin = module.input_data()   # loads input.json
    jout = main(jin)
    module.write_output_data(jout)  # writes output.json




from allensdk.api.queries.biophysical_api import \
    BiophysicalApi
from allensdk.api.queries.cell_types_api import CellTypesApi
from allensdk.api.queries.rma_api import RmaApi
import os, sys, shutil


#bp = BiophysicalApi('http://api.brain-map.org')
#bp.cache_stimulus = True # change to False to not download the large stimulus NWB file
# neuronal_model_id = 472451419    # get this from the web site as above
# bp.cache_data(neuronal_model_id, working_directory='neuronal_model')

# According to this there are 49 biophysical models.
# http://api.brain-map.org/api/v2/data/query.json?q=model::NeuronalModel,rma::include,neuronal_model_template[name$eq%27Biophysical%20-%20perisomatic%27],rma::options[num_rows$eqall]

# Note, am I supposed to be only archiving Biophysical models or also GLIFs?

class BiophysicalArchiver(object):
    def __init__(self, archive_dir=None):
        self.bp = BiophysicalApi('http://api.brain-map.org')
        self.bp.cache_stimulus = True # change to False to not download the large stimulus NWB file
        self.cta = CellTypesApi()
        self.rma = RmaApi()
        self.neuronal_model_download_endpoint = 'http://celltypes.brain-map.org/neuronal_model/download/'
        self.template_names = {}
        self.nwb_list = []
        
        if archive_dir == None:
            archive_dir = '.'
        self.archive_dir = archive_dir
    
    def get_template_names(self):
        template_response = self.rma.model_query('NeuronalModelTemplate')
        self.template_names = { t['id']: str(t['name']).replace(' ', '_') for t in template_response}
    
    def get_cells(self):
        return self.cta.list_cells(True, True)
    
    def get_neuronal_models(self, specimen_ids):
        return self.rma.model_query('NeuronalModel',
                                    criteria='specimen[id$in%s]' % ','.join(str(i) for i in specimen_ids),
                                    include='specimen',
                                    num_rows='all')
    
    def get_stimulus_file(self, neuronal_model_id):
        result = self.rma.model_query('NeuronalModel',
                                      criteria='[id$eq%d]' % (neuronal_model_id),
                                      include="specimen(ephys_result(well_known_files(well_known_file_type[name$il'NWB*'])))",
                                      tabular=['path'])
        
        stimulus_filename = result[0]['path']
        
        return stimulus_filename
        
        
        stimulus_filename = os.path.basename(result[0]['path'])
        
        return stimulus_filename
    
    def archive_cell(self, ephys_result_id, specimen_id, template, neuronal_model_id):
        url = self.neuronal_model_download_endpoint + "/%d" % (neuronal_model_id)
        file_name = os.path.join(self.archive_dir, 'ephys_result_%d_specimen_%d_%s_neuronal_model_%d.zip' % (ephys_result_id,
                                                                                                             specimen_id,
                                                                                                             template,
                                                                                                             neuronal_model_id))
        self.rma.retrieve_file_over_http(url, file_name)
        nwb_file = self.get_stimulus_file(neuronal_model_id)
        shutil.copy(nwb_file, self.archive_dir) 
        self.nwb_list.append("%s\t%s" % (os.path.basename(nwb_file),
                                         file_name))
    
if __name__ == '__main__':
    archive_dir = sys.argv[-1] # /data/informatics/mousecelltypes/model_cache_may_2015
    ba = BiophysicalArchiver(archive_dir)
    ba.get_template_names()
    cells = ba.get_cells()
    
    specimen_ids = (cell['id'] for cell in cells)
    neuronal_models = ba.get_neuronal_models(specimen_ids)
    for nm in neuronal_models:
        ephys_result_id = nm['specimen']['ephys_result_id']
        template_id = nm['neuronal_model_template_id'] 
        if template_id in ba.template_names:
            template = ba.template_names[template_id]
        else:
            template = 'unknown'
        ba.archive_cell(ephys_result_id, nm['specimen_id'], template, nm['id'])
    with open(os.path.join(ba.archive_dir, 'STIMULUS.csv'), 'w') as f:
        f.write("\n".join(ba.nwb_list))


import logging

import numpy as np


class Tile(object):

    def __init__(self, index, image, is_missing, bounds, channel, size, margins, *args, **kwargs):

        # identifier
        self.index = index

        # actual image data
        self.image = image
        self.is_missing = is_missing

        # parameters related to the position of the tile within a larger image
        self.bounds = bounds
        self.channel = channel

        # parameters related to the valid portion of the tile
        self.size = size
        self.margins = margins

        logging.info('tile {index} on channel {channel} starts at ({0}, {1})'.format(self.bounds['row']['start'], 
                                                                                     self.bounds['column']['start'], 
                                                                                     index=self.index, 
                                                                                     channel=self.channel))


    def trim_self(self):
        logging.info('trimming tile')
        self.image = self.trim(self.image)

    
    def trim(self, image):
        logging.info('trimming with margins ({row}, {column})'.format(**self.margins))

        return image[self.margins['row']: self.margins['row'] + self.size['row'], 
                     self.margins['column']: self.margins['column'] + self.size['column']]


    def average_tile_is_untrimmed(self, average_tile):
        return average_tile.shape[0] > self.image.shape[0] \
            or average_tile.shape[1] > self.image.shape[1]


    def apply_average_tile(self, average_tile):

        if average_tile is None:
            logging.info('no average tile found for tile with index {index} on channel {channel}'.format(**self.__dict__))
            return self.image

        if self.average_tile_is_untrimmed(average_tile):
            logging.info('trimming average tile')
            average_tile = self.trim(average_tile)

        logging.info('applying flatfield correction to tile with index {index} on channel {channel}'.format(**self.__dict__))
        return np.multiply(self.image, average_tile)

    
    def apply_average_tile_to_self(self, average_tile):
        self.image = self.apply_average_tile(average_tile)


    def get_image_region(self):
        
        row = self.bounds['row']
        col = self.bounds['column']

        return [slice(row['start'], row['end']), 
                slice(col['start'], col['end']), 
                self.channel]


    def get_missing_path(self):

        row = self.bounds['row']
        col = self.bounds['column']
        
        path = [row['start'], col['start'], 
                row['end'], col['start'], 
                row['end'], col['end'], 
                row['start'], col['end']]

        logging.info('missing tile starts at: ({0}, {1})'.format(*path))
        return path


    def initialize_image(self):
        logging.info('initializing tile image to 0')
        self.image = np.zeros((self.size['row'], self.size['column']))



from allensdk.model.biophys_sim.neuron.hoc_utils import HocUtils
from allensdk.ephys.ephys_extractor import EphysSweepFeatureExtractor
import logging

import numpy as np

class Utils(HocUtils):
    _log = logging.getLogger(__name__)

    def __init__(self, description):
        super(Utils, self).__init__(description)
        self.stim = None
        self.stim_curr = None
        self.sampling_rate = None
        self.cell = self.h.cell()

    def generate_morphology(self, morph_filename):
        h = self.h
        cell = self.cell

        swc = self.h.Import3d_SWC_read()
        swc.quiet = 1
        swc.input(str(morph_filename))
        imprt = self.h.Import3d_GUI(swc, 0)
        imprt.instantiate(cell)

        for seg in cell.soma[0]:
            seg.area()

        for sec in cell.all:
            sec.nseg = 1 + 2 * int(sec.L / 40)

        cell.simplify_axon()
        for sec in cell.axonal:
            sec.L = 30
            sec.diam = 1
            sec.nseg = 1 + 2 * int(sec.L / 40)
        cell.axon[0].connect(cell.soma[0], 0.5, 0)
        cell.axon[1].connect(cell.axon[0], 1, 0)
        h.define_shape()

    def load_cell_parameters(self):
        cell = self.cell
        passive = self.description.data['passive'][0]
        conditions = self.description.data['conditions'][0]
        channels = self.description.data['channels']
        addl_params = self.description.data['addl_params']

        # Set passive properties
        for sec in cell.all:
            sec.Ra = passive['ra']
            sec.cm = passive['cm'][sec.name().split(".")[1][:4]]
            sec.insert('pas')
            for seg in sec:
                seg.pas.e = passive["e_pas"]
        self.h.v_init = passive["e_pas"]

        # Insert channels and set parameters
        for c in channels:
            if c["mechanism"] != "":
                sections = [s for s in cell.all if s.name().split(".")[1][:4] == c["section"]]
                for sec in sections:
                    sec.insert(c["mechanism"])

        for ap in addl_params:
            if ap["mechanism"] != "":
                sections = [s for s in cell.all if s.name().split(".")[1][:4] == ap["section"]]
                for sec in sections:
                    sec.insert(ap["mechanism"])

        # Set reversal potentials
        for erev in conditions['erev']:
            sections = [s for s in cell.all if s.name().split(".")[1][:4] == erev["section"]]
            for sec in sections:
                sec.ena = erev["ena"]
                sec.ek = erev["ek"]

    def set_normalized_parameters(self, params):
        channels_and_others = self.description.data['channels'] + self.description.data['addl_params']
        for i, p in enumerate(params):
            c = channels_and_others[i]
            value = p * (c["max"] - c["min"]) + c["min"]
            sections = [s for s in self.cell.all if s.name().split(".")[1][:4] == c["section"]]
            for sec in sections:
                param_name = c["parameter"]
                if c["mechanism"] != "":
                    param_name += "_" + c["mechanism"]
                setattr(sec, param_name, value)

    def set_actual_parameters(self, params):
        channels_and_others = self.description.data['channels'] + self.description.data['addl_params']
        for i, p in enumerate(params):
            c = channels_and_others[i]
            sections = [s for s in self.cell.all if s.name().split(".")[1][:4] == c["section"]]
            for sec in sections:
                param_name = c["parameter"]
                if c["mechanism"] != "":
                    param_name += "_" + c["mechanism"]
                setattr(sec, param_name, p)

    def normalize_actual_parameters(self, params):
        params_array = np.array(params)
        channels_and_others = self.description.data['channels'] + self.description.data['addl_params']
        max_vals = np.array([c["max"] for c in channels_and_others])
        min_vals = np.array([c["min"] for c in channels_and_others])

        normalized_params = (params_array - min_vals) / (max_vals - min_vals)
        return normalized_params.tolist()

    def actual_parameters_from_normalized(self, params):
        actual_params = []
        channels_and_others = self.description.data['channels'] + self.description.data['addl_params']
        for i, p in enumerate(params):
            c = channels_and_others[i]
            value = p * (c["max"] - c["min"]) + c["min"]
            actual_params.append(value)
        return actual_params

    def insert_iclamp(self):
        self.stim = self.h.IClamp(self.cell.soma[0](0.5))

    def set_iclamp_params(self, amp, delay, dur):
        self.stim.amp = amp
        self.stim.delay = delay
        self.stim.dur = dur

    def calculate_feature_errors(self, t_ms, v, i):
        # Special case checks and penalty values
        minimum_num_spikes = 2
        missing_penalty_value = 20.0
        max_fail_penalty = 250.0
        min_fail_penalty = 75.0
        overkill_reduction = 0.75
        variance_factor = 0.1

        fail_trace = False

        delay = self.stim.delay * 1e-3
        duration = self.stim.dur * 1e-3
        t = t_ms * 1e-3
        feature_names = self.description.data['features']

        # penalize for failing to return to rest
        start_index = np.flatnonzero(t >= delay)[0]
        if np.abs(v[-1] - v[:start_index].mean()) > 2.0:
            fail_trace = True
        else:
            swp = EphysSweepFeatureExtractor(t, v, i, start=0, end=delay, filter=None)
            swp.process_spikes()
            if swp.sweep_feature("avg_rate") > 0:
                fail_trace = True

        target_features = self.description.data['target_features']
        target_features_dict = {f["name"]: {"mean": f["mean"], "stdev": f["stdev"]} for f in target_features}

        if not fail_trace:
            swp = EphysSweepFeatureExtractor(t, v, i, start=delay, end=(delay + duration), filter=None)
            swp.process_spikes()
            if len(swp.spikes()) < minimum_num_spikes: # Enough spikes?
                fail_trace = True
            else:
                avg_per_spike_peak_error = np.mean([abs(spk["peak_v"] - target_features_dict["peak_v"]["mean"]) for spk in swp.spikes()])
                avg_overall_error = abs(target_features_dict["peak_v"]["mean"] - swp.spike_feature("peak_v").mean())
                if avg_per_spike_peak_error > 3.0 * avg_overall_error: # Weird bi-modality of spikes; 3.0 is arbitrary
                    fail_trace = True

        if fail_trace:
            variance_start = np.flatnonzero(t >= delay - 0.1)[0]
            variance_end = np.flatnonzero(t >= (delay + duration) / 2.0)[0]
            trace_variance = v[variance_start:variance_end].var()
            error_value = max(max_fail_penalty - trace_variance * variance_factor, min_fail_penalty)
            errs = np.ones(len(feature_names)) * error_value
        else:
            errs = []

            # Calculate additional features not done by swp.process_spikes()
            baseline_v = swp.sweep_feature("v_baseline")
            other_features = {}
            threshold_t = swp.spike_feature("threshold_t")
            fast_trough_t = swp.spike_feature("fast_trough_t")
            slow_trough_t = swp.spike_feature("slow_trough_t")

            delta_t = slow_trough_t - fast_trough_t
            delta_t[np.isnan(delta_t)] = 0.
            other_features["slow_trough_delta_time"] = np.mean(delta_t[:-1] / np.diff(threshold_t))

            fast_trough_v = swp.spike_feature("fast_trough_v")
            slow_trough_v = swp.spike_feature("slow_trough_v")
            delta_v = fast_trough_v - slow_trough_v
            delta_v[np.isnan(delta_v)] = 0.
            other_features["slow_trough_delta_v"] = delta_v.mean()

            for f in feature_names:
                target_mean = target_features_dict[f]['mean']
                target_stdev = target_features_dict[f]['stdev']

                if target_stdev == 0:
                    print("Feature with 0 stdev: ", f)

                if f in swp.spike_feature_keys():
                    model_mean = swp.spike_feature(f).mean()
                elif f in swp.sweep_feature_keys():
                    model_mean = swp.sweep_feature(f)
                elif f in other_features:
                    model_mean = other_features[f]
                else:
                    model_mean = np.nan

                if np.isnan(model_mean):
                    errs.append(missing_penalty_value)
                else:
                    errs.append(np.abs((model_mean - target_mean) / target_stdev))

            errs = np.array(errs)
        return errs

    def record_values(self):
        v_vec = self.h.Vector()
        t_vec = self.h.Vector()
        i_vec = self.h.Vector()

        v_vec.record(self.cell.soma[0](0.5)._ref_v)
        i_vec.record(self.stim._ref_amp)
        t_vec.record(self.h._ref_t)

        return v_vec, i_vec, t_vec



import sys
import string

class ResourceFile(object):
    def __init__(self):
        self.key_value = {}
        self.accessed = {}

    def load(self, infile):
        """ Reads in a json, yaml or toml resource file. If multiple files
            are loaded, resources between them are merged. An error 
            occurs if the same key is loaded multiple times and different
            values are defined for it
        """
        if infile.endswith("json"):
            self.load_json(infile)
        elif infile.endswith("yml") or infile.endswith("yaml"):
            self.load_yaml(infile)
        elif infile.endswith("tml") or infile.endswith("toml"):
            self.load_toml(infile)
        else:
            print "Unrecognized extension for file '%s'. Please use json, yaml or toml" % infile
            sys.exit(1)
    
    def get(self, key, default=None, replace_table=None):
        """ Returns the resource value associated with the provided key.

            Arguments:
                *key* (text) Name of resource

                *default* (text) Value to be returned if key isn't found

                *replace_table* (dict) Substrings that are to be replaced
                in resource string. E.g., if replace_table = { foo: "bar" }
                then all instances of "foo" in the resource string will be
                replaced with "bar"

            Returns:
                Resource string if found and default value if not, with
                applied substitutions from replace_table
        """
        self.accessed[key] = True
        val = self.key_value.get(key, default)
        if replace_table is not None:
            for k,v in replace_table.iteritems():
                val = string.replace(val, k, v)
        return val

    def report(self):
        """ Reports all resources that were defined but not used
        """
        err = False
        print "-------------------------------"
        print "---     Resource report      --"
        for k,v in self.accessed.iteritems():
            if not v:
                if not err:
                    err = True
                print "\t'%s' not used" % k
        if not err:
            print "all defined resources were used"
        print "-------------------------------"

    ####################################################################
    # internal procedure to load json file
    def load_json(self, infile):
        """ Reads input json, yaml or toml file
        """
        import json
        try:
            with open(infile, 'r') as f:
                resources = json.load(f)
                f.close()
        except IOError:
            print "Unable to open input json file '%s'" % infile
            sys.exit(1)
        self.read_keys(resources)

    # internal procedure to load yaml
    def load_yaml(self, infile):
        try:
            import yaml
            try:
                with open(infile, 'r') as f:
                    resources = yaml.load(f)
                    f.close()
            except IOError:
                print "Unable to open input yaml file '%s'" % infile
                sys.exit(1)
            self.read_keys(resources)
        except ImportError:
            print "*** yaml not available -- please pip install pyyaml"
            sys.exit(1)

    # internal procedure to load toml
    def load_toml(self, infile):
        try:
            import toml
            try:
                with open(infile, 'r') as f:
                    resources = toml.load(f)
                    f.close()
            except IOError:
                print "Unable to open input toml file '%s'" % infile
                sys.exit(1)
            self.read_keys(resources)
        except ImportError:
            print "*** toml not available -- please pip install toml"
            sys.exit(1)


    # internal procedure to read keys out of dictionary recursively
    def read_keys(self, resources):
        err = False
        for k,v in resources.iteritems():
            if isinstance(v, dict):
                self.read_keys(v)
            else:
                if k in self.key_value and v != self.key_value[k]:
                    print "Error -- inconsistent values for key '%s'" % k
                    err = True
                self.key_value[k] = v
                self.accessed[k] = False
        if err:
            sys.exit(1)


class ResourceFileTest(object):
    def create_json(self, fname, rsrc):
        import json
        d = {}
        d["jone"] = "json one"
        d["jtwo"] = "json one"
        d["jsub"] = {}
        d["jsub"]["jthree"] = "json three"
        with open(fname, 'w') as f:
            json.dump(d, f, indent=2)
        f.close()
        rsrc.load(fname)

    def create_yaml(self, fname, rsrc):
        try:
            import yaml
            d = {}
            d["yone"] = "yaml one"
            d["ytwo"] = "yaml one"
            d["ysub"] = {}
            d["ysub"]["ythree"] = "yaml three"
            with open(fname, 'w') as f:
                yaml.dump(d, f, indent=2)
            f.close()
            rsrc.load(fname)
        except ImportError:
            print "*** yaml not available -- please pip install pyyaml"

    def create_toml(self, fname, rsrc):
        try:
            import toml
            d = {}
            d["tone"] = "toml one"
            d["ttwo"] = "toml one"
            d["tsub"] = {}
            d["tsub"]["tthree"] = "toml three"
            with open(fname, 'w') as f:
                toml.dump(d, f)
            f.close()
            rsrc.load(fname)
        except ImportError:
            print "*** toml not available -- please pip install toml"

    def run(self):
        # create and load json, yaml and toml files
        rsrc = ResourceFile()
        self.create_json("tmp.json", rsrc)
        self.create_yaml("tmp.yaml", rsrc)
        self.create_toml("tmp.toml", rsrc)
        # print resource from each
        print rsrc.get("jone", "json error")
        print rsrc.get("yone", "** yaml error")
        print rsrc.get("tone", "** toml error")
        print rsrc.get("jthree", "json error")
        print rsrc.get("ythree", "** yaml error")
        print rsrc.get("tthree", "** toml error")
        # run report
        rsrc.report()




import numpy as np
from collections import Counter
from allensdk.ephys.feature_extractor import EphysFeatureExtractor
import allensdk.internal.model.biophysical.ephys_utils as ephys_utils

def calculate_fi_curves(data_set, sweeps):

    sweep_type = "C1LSCOARSE"
    _, sweep_numbers, statuses = ephys_utils.get_sweeps_of_type(sweep_type, sweeps)
    features = EphysFeatureExtractor()

    coarse_fi_curve = []
    sweep_status = dict(zip(sweep_numbers, statuses))

    for s in sweep_numbers:
        if sweep_status[s] in [ 'auto_failed', 'manual_failed' ]:
            continue

        v, i, t = ephys_utils.get_sweep_v_i_t_from_set(data_set, s)
        if np.all(v[-100:] == 0):
            continue
        stim_start, stim_dur, stim_amp, start_idx, end_idx = ephys_utils.get_step_stim_characteristics(i, t)
        features.process_instance(s, v, i, t, stim_start, stim_dur, "")
        coarse_fi_curve.append((stim_amp, features.feature_list[-1].mean["n_spikes"] / stim_dur))

    sweep_type = "C2SQRHELNG"
    core2_fi_curve = []
    core2_half_fi_curve = []
    _, sweep_numbers, statuses = ephys_utils.get_sweeps_of_type(sweep_type, sweeps)
    sweep_status = dict(zip(sweep_numbers, statuses))
    core2_amps = {}
    amp_list = []
    for s in sweep_numbers:
        if sweep_status[s] in [ 'auto_failed', 'manual_failed' ]:
            continue

        v, i, t = ephys_utils.get_sweep_v_i_t_from_set(data_set, s)
        if np.all(v[-100:] == 0):
            continue
        stim_start, stim_dur, stim_amp, start_idx, end_idx = ephys_utils.get_step_stim_characteristics(i, t)
        if stim_start is np.nan:
            sweep_status[s] = "manual_failed"
            continue
        core2_amps[s] = stim_amp
        amp_list.append(stim_amp)

    core2_amp_counter = Counter(amp_list)
    common_amps = core2_amp_counter.most_common(3)

    features = EphysFeatureExtractor()
    for amp, count in common_amps:
        for k in core2_amps:
            if core2_amps[k] == amp and sweep_status[k][-6:] == "passed":
                v, i, t = ephys_utils.get_sweep_v_i_t_from_set(data_set, k)
                stim_start, stim_dur, stim_amp, start_idx, end_idx = ephys_utils.get_step_stim_characteristics(i, t)
                features.process_instance(s, v, i, t, stim_start, stim_dur, "")
                core2_fi_curve.append((amp, features.feature_list[-1].mean["n_spikes"] / stim_dur))
                first_half_spike_count = len([spk for spk in features.feature_list[-1].mean["spikes"] if spk["t"] < stim_start + stim_dur / 2.0])
                core2_half_fi_curve.append((amp, first_half_spike_count / (stim_dur / 2.0)))

    return { "coarse": coarse_fi_curve, "core2": core2_fi_curve, "core2_half": core2_half_fi_curve }

def estimate_fi_shift(data_set, sweeps):
    curve_data = calculate_fi_curves(data_set, sweeps)

    # Linear fit to original fI curve
    coarse_fi_sorted = sorted(curve_data["coarse"], key=lambda d: d[0])
    x = np.array([d[0] for d in coarse_fi_sorted], dtype=np.float64)
    y = np.array([d[1] for d in coarse_fi_sorted], dtype=np.float64)

    if len(np.flatnonzero(y)) == 0: # original curve is all zero, so can't figure out shift
        return np.nan, 0

    last_zero_index = np.flatnonzero(y)[0] - 1
    A = np.vstack([x[last_zero_index:], np.ones(len(x[last_zero_index:]))]).T
    m, c = np.linalg.lstsq(A, y[last_zero_index:])[0]

    # Relative error of later traces to best-fit line
    if len(curve_data["core2_half"]) < 1:
        return np.nan, 0
    # FIX TO RECTIFY PREDICTED FI CURVE
    x_shift = [amp - (freq - c) / m for amp, freq in curve_data["core2_half"]]
    return np.mean(x_shift), len(x_shift)



import itertools
import cPickle
import logging
from allensdk.internal.brain_observatory import roi_filter_utils
import allensdk.internal.brain_observatory.mask_set as mask_set
from allensdk.brain_observatory.roi_masks import create_roi_mask_array
from sklearn.cross_validation import cross_val_score
from sklearn import __version__ as sklearn_version
import numpy as np
import pandas as pd


class ROIClassifier(object):
    '''Wrapper for machine learning classifier.

    Provides an underlying classifier model implementing `fit`,
    `score`, and `predict`. Tracks additional information for
    constructing the feature array from input datastreams, as well
    as training data used and cross validation scores generated.

    Parameters
    ----------
    model_data : dictionary
        Dictionary of classifier properties
        `sklearn_version`: Version of sklearn used for training.
        `model`: Underlying classifier.
        `training_features`: Feature set used to train model.
        `training_labels`: Label set used to train model.
        `trimmed_features`: Features to remove from input data.
        `structure_ids`: Structure ID set used for training.
        `drivers`: Driver set used for training.
        `reporters`: Reporter set used for training.
        `other_appended_labels`: Labels appended outside model.
        `cross_validation_scores`: Cross validation if generated.
    '''
    def __init__(self, model_data=None):
        '''Constructor.'''
        if model_data is None:
            model_data = {}
        self.sklearn_version = sklearn_version
        model_sklearn = model_data.get("sklearn_version", None)
        if sklearn_version != model_sklearn:
            logging.warning("Using sklearn %s, model trained using %s",
                            sklearn_version, model_sklearn)
        self.model = model_data.get("model", None)
        self.training_features = model_data.get("training_features",
                                                pd.DataFrame())
        self.training_labels = model_data.get("training_labels",
                                              pd.DataFrame())
        self.trimmed_features = model_data.get("trimmed_features", [])
        self.structure_ids = model_data.get("structure_ids", [])
        self.drivers = model_data.get("drivers", [])
        self.reporters = model_data.get("reporters", [])
        self.other_appended_labels = model_data.get("other_appended_labels",
                                                    [])
        # this is a harsh score for multilabel because it requires ALL
        # labels predicted
        self.cross_validation_scores = model_data.get(
            "cross_validation_scores", None)
        self.unexpected_features = []

    @property
    def model_data(self):
        '''The classifier properties as a dictionary.'''
        data = {"model": self.model,
                "training_features": self.training_features,
                "training_labels": self.training_labels,
                "trimmed_features": self.trimmed_features,
                "structure_ids": self.structure_ids,
                "drivers": self.drivers,
                "reporters": self.reporters,
                "other_appended_labels": self.other_appended_labels,
                "sklearn_version": self.sklearn_version,
                "cross_validation_scores": self.cross_validation_scores}
        return data

    @property
    def label_names(self):
        '''Return label names for the classifier.'''
        return self.training_labels.columns

    def create_feature_array(self, object_data, depth, structure_id, drivers,
                             reporters):
        '''Creates feature array from input data.

        See Also
        --------
        create_feature_array : Create a feature array given model and inputs
        '''
        features = create_feature_array(self.model_data, object_data, depth,
                                        structure_id, drivers, reporters)

    def get_labels(self, object_data, depth, structure_id, drivers,
                   reporters):
        '''Generate labels from input data.

        See Also
        --------
        ROIClassifier.create_feature_array
        '''
        features = create_feature_array(self.model_data, object_data, depth,
                                        structure_id, drivers, reporters)
        self.unexpected_features = get_unexpected_features(
            self.model_data, object_data, structure_id, drivers, reporters)
        return self.predict(features)

    def fit(self, features, labels):
        '''Fit model to data.

        Parameters
        ----------
        features : pandas.DataFrame
            Training feature set.
        labels : pandas.DataFrame
            Training labels.
        '''
        self.training_features = features
        self.training_labels = labels
        self.model.fit(features, labels)

    def score(self, features, labels):
        '''Calculate classifier score on data.'''
        return self.model.score(features, labels)

    def predict(self, features):
        '''Generate classification labels given features.'''
        return self.model.predict(features)

    def cross_validate(self, features, labels, n_folds=5, n_jobs=1):
        '''Generate cross-validation scores for the classifier.

        Parameters
        ----------
        features : pandas.DataFrame
            Set of features for classification.
        labels : pandas.DataFrame
            Set of ground truth labels for training and evaluation.
        n_folds : int
            Number of folds for K-Fold cross-validation.
        n_jobjs : int
            Number of CPUs to use.

        Returns
        -------
        numpy.ndarray
            `n_folds` cross-validation scores.
        '''
        self.cross_validation_scores = cross_val_score(
            self.model, features, labels, cv=n_folds, n_jobs=n_jobs)
        return self.cross_validation_scores

    def save(self, filename):
        '''Save the classifier to file by pickling.'''
        with open(filename, "wb") as f:
            cPickle.dump(self.model_data, f)

    @staticmethod
    def from_file(filename):
        '''Load an ROIClassifier from file.'''
        with open(filename, "rb") as f:
            return ROIClassifier(cPickle.load(f))


def mean_gray_to_sigma(meanInt0, snpoffsetstdv):
    '''Calculate intensity variation used in prior code.

    Parameters
    ----------
    meanInt0 : pandas.Series
        Array of intensity averages.
    snpoffsetstdv : pandas.Series
        Array of soma-neuropil standard deviations.

    Returns
    -------
    pandas.Series
        meanInt0/snpoffsetstdv, preventing Inf (returns as 0).
    '''
    mean_gray_to_sigma = meanInt0 / snpoffsetstdv.astype(float)
    mean_gray_to_sigma[snpoffsetstdv == 0.0] = 0
    return mean_gray_to_sigma


def create_feature_array(model_data, object_data, depth, structure_id,
                         drivers, reporters):
    '''Create feature array from input data.

    This creates the feature array with column ordering matching what
    the classifier was trained on.

    Parameters
    ----------
    model_data : dictionary
        Dictionary containing information about the machine learning
        model and training set.
    object_data : pandas.DataFrame
        Object list data.
    depth : float
        Imaging depth of the experiment.
    structure_id : string
        Targeted structure id.
    drivers : list
        List of drivers for the mouse.
    reporters : list
        List of reporters for the mouse.
    '''
    training_features = model_data["training_features"].columns
    if np.isnan(depth):
        depth = 0
    meanGrayToSigma = mean_gray_to_sigma(
        object_data["meanInt0"], object_data["snpoffsetstdv"])
    features = pd.DataFrame()
    for column in training_features:
        if column == "depth":
            features[column] = depth
        # special case that isn't in object list
        elif column == "meanGrayToSigma":
            features[column] = meanGrayToSigma
        elif column in model_data["structure_ids"]:
            features[column] = int(structure_id == column)
        elif column in model_data["drivers"]:
            features[column] = int(column in drivers)
        elif column in model_data["reporters"]:
            features[column] = int(column in reporters)
        elif column in object_data.columns:
            features[column] = object_data[column]
        else:
            logging.error("Feature %s missing from input data", column)
            raise KeyError(
                "Feature {} missing from input data".format(column))
    return features


def get_unexpected_features(model_data, object_data, structure_id, drivers,
                            reporters):
    '''Get list of incoming features that weren't in traning data.

    Parameters
    ----------
    model_data : dictionary
        Dictionary containing information about the machine learning
        model and training set.
    object_data : pandas.DataFrame
        Object list data.
    structure_id : string
        Targeted structure id.
    drivers : list
        List of drivers for the mouse.
    reporters : list
        List of reporters for the mouse.
    '''
    training_features = model_data["training_features"].columns
    trimmed_features = model_data["trimmed_features"]
    inputs = list(itertools.chain(object_data.columns, [structure_id],
                                  drivers, reporters))
    unexpected_features = []
    for feature in inputs:
        if (feature not in training_features) and \
           (feature not in trimmed_features):
            unexpected_features.append(feature)
    return unexpected_features


def label_unions_and_duplicates(rois, overlap_threshold):
    '''Detect unions and duplicates and label ROIs.'''
    masks = create_roi_mask_array(rois)
    valid_masks = np.ones(masks.shape[0]).astype(bool)
    ms = mask_set.MaskSet(masks=masks)

    # detect and label duplicates
    duplicates = ms.detect_duplicates(overlap_threshold)
    for duplicate in duplicates:
        index = duplicate[0]
        if "duplicate" not in rois[index].labels:
            rois[index].labels.append("duplicate")
        valid_masks[index] = False

    # detect and label unions only for remaining valid masks
    valid_idxs = np.where(valid_masks)
    ms = mask_set.MaskSet(masks=masks[valid_idxs].astype(bool))
    unions = ms.detect_unions()

    if unions:
        union_idxs = unions.keys()
        idxs = valid_idxs[0][union_idxs]
        for idx in idxs:
            if "union" not in rois[idx].labels:
                rois[idx].labels.append("union")
    return rois


def apply_labels(rois, label_array, label_names):
    '''Apply labels to rois.

    Parameters
    ----------
    rois : list
        List of RoiMask objects sorted to `label_array` order.
    label_array : numpy.ndarray
        Label array output from classifier.
    label_names : list
        Names to apply to columns of `label_array`.

    Returns
    -------
    list
        List of ROIs with labels appended.
    '''
    label_df = pd.DataFrame(data=label_array, columns=label_names)
    label_lists = label_df.apply(_column_match).apply(
        _compress_to_list, args=(label_df.columns,), axis=1)
    for i, roi in enumerate(rois):
        roi.labels.extend(label_lists[i])
    return rois


def _column_match(column):
    return column == 1


def _compress_to_list(row, names):
    '''Get names that have value 1 in row.'''
    return list(names[row.values])



from __future__ import division
import logging
import sys
import functools

import numpy as np
from scipy.ndimage.interpolation import zoom
from six import iteritems

sys.path.append('/shared/bioapps/itk/itk_shared/jp2/build')
import jpeg_twok

import image_utilities as iu
from base_subimage import PolygonSubImage, SegmentationSubImage, IntensitySubImage


#==============================================================================


class CavSubImage(PolygonSubImage):

    required_polys = ['missing_tile', 'cav_tracer']


    def compute_coarse_planes(self):

        nonmissing = np.logical_not(self.images['missing_tile'])
        del self.images['missing_tile']

        self.apply_pixel_counter('sum_pixels', nonmissing)

        cav_nonmissing = np.multiply(self.images['cav_tracer'], nonmissing)
        del nonmissing
        self.apply_pixel_counter('cav_tracer', cav_nonmissing)
        del cav_nonmissing

        del self.images


#==============================================================================



import logging

import numpy as np

import time

from scipy.optimize import fminbound, fmin
from scipy.optimize import minimize

import json

from uuid import uuid4

import allensdk.internal.model.glif.error_functions as error_functions

# TODO: clean up
# TODO: license
# TODO: document

class GlifOptimizer(object):
    def __init__(self, experiment, dt, 
                 outer_iterations, inner_iterations, 
                 sigma_outer, sigma_inner,
                 param_fit_names, stim,                 
                 xtol, ftol, 
                 internal_iterations, 
                 bessel,
                 error_function = None,
                 error_function_data = None,
                 init_params = None):

        self.start_time = None
        self.rng = np.random.RandomState()

        self.experiment = experiment
        self.dt = dt
        self.outer_iterations = outer_iterations
        self.inner_iterations = inner_iterations
        self.init_params = init_params
        self.sigma_outer = sigma_outer
        self.sigma_inner = sigma_inner
        self.param_fit_names = param_fit_names
        self.stim = stim

        # use MLIN by default
        if error_function is None:
            error_function = error_functions.MLIN_list_error

        self.error_function = error_function
        self.error_function_data = error_function_data

        self.xtol = xtol
        self.ftol = ftol

        self.internal_iterations = internal_iterations
        
        self.bessel = bessel

        logging.info('internal_iterations: %s' % internal_iterations)
        logging.info('outer_iterations: %s' % outer_iterations)
        logging.info('inner_iterations: %s' % inner_iterations)

        self.iteration_info = [];

        expected_param_count = experiment.neuron_parameter_count()

        if self.init_params is None:
            self.init_params = np.ones(expected_param_count)
        elif len(self.init_params) != expected_param_count:
            self.init_params = np.ones(expected_param_count)
            logging.warning('optimizer init_params has wrong length (given %d, expected %d). settings to all ones' % (len(self.init_params), expected_param_count))
                          
    def to_dict(self):
        return {
            'outer_iterations': self.outer_iterations,
            'inner_iterations': self.inner_iterations,
            'init_params': self.init_params,
            'sigma_outer': self.sigma_outer,
            'sigma_inner': self.sigma_inner,
            'param_fit_names': self.param_fit_names,
            'xtol': self.xtol,
            'ftol': self.ftol,
            'internal_iterations': self.internal_iterations,
            'iteration_info': self.iteration_info,
            'bessel': self.bessel
        }
            
    def randomize_parameter_values(self, values, sigma):
        values = np.array(self.rng.normal(values, sigma))

        # values might not have a shape if it's a single element long, depending on your numpy version
        if not values.shape:
            values = np.array([values])
        return values
    
    def initiate_unique_seed(self, seed=None):
        
        if seed == None:
            x1=str(int(uuid4())) #get a uuid, turn it into int then turn it into string
            x2=[x1[ii:ii+8]for ii in range(0,40,8)] #break it up into chunks    
            x3=[int(ii) for ii in x2]#turn string chunks back into integers
            print 'seed', x3
            self.rng.seed(x3)
        else:
            self.rng.seed(seed)

    def evaluate(self, x, dt_multiplier=100):

        self.experiment.neuron.dt_multiplier = dt_multiplier
        return self.error_function([x], self.experiment, self.error_function_data)
        
    def run_many(self, iteration_finished_callback=None, seed=None):
        self.initiate_unique_seed(seed=seed)
        params_start = self.init_params
        self.start_time = time.time()
        params=params_start
#        params=self.randomize_parameter_values(params_start, self.sigma_outer)
        print 'actual starting parameters', params
        
        stop_flag=False
        
        # TODO: unhardcode this
        dt_multiplier_list = [100, 32, 10]
        #Note the following line may be useful when there are more iteration but is hasnt been tested
#        dt_multiplier_list = np.ceil(np.logspace(1,2,self.inner_iterations))[::-1].astype(int)
        print dt_multiplier_list
#         dt_multiplier_list = [10,10,10]
        #TODO: figure out the implications of this being an int versus float
        #TODO: make this so that dt multiplier actually gets set
        for outer in range(0, self.outer_iterations):  #outerloop 
            for inner in range(0, self.inner_iterations):  #innerloop
                iteration_start_time = time.time()

                # run the optimizer once.  first time is always the passed initial conditions.
#                 print 'dt_multiplier_list[inner]', dt_multiplier_list[inner]
                #--set this equal to 1 if want to do it slow
                self.experiment.neuron.dt_multiplier = dt_multiplier_list[inner]
                #self.experiment.neuron.dt_multiplier = 10
                
                
                opt = self.run_once(params)
                xopt, fopt = opt[0], opt[1]

                logging.info('fmin took %f secs, %f mins, %f hours' %  (time.time() - iteration_start_time, (time.time() - iteration_start_time)/60, (time.time() - iteration_start_time)/60/60))

                self.iteration_info.append({
                    'in_params': np.array(params).tolist(),
                    'out_params': xopt.tolist(),
                    'error': float(fopt),
                    'dt_multiplier': self.experiment.neuron.dt_multiplier
                })

#                ER=self.iteration_info['error']
#                ETOL=1.e-4
#                if len(ER) >=3:
#                    #!!!!!!!!!!!!!!!fix this to use that actual parameters!!!!!!!!!!!!!!!!!!!!!!
#                    if np.abs(ER[-1]-ER[-2])<ETOL and np.abs(ER[-1]-ER[-3])<ETOL and np.abs(ER[-2]-ER[-3])<ETOL:
#                        stop_flag=True
                    
                    

                if iteration_finished_callback is not None:
                    iteration_finished_callback(self, outer, inner)
                    
                if stop_flag is True:
                    break

                # randomize the best fit parameters
                params = self.randomize_parameter_values(xopt, self.sigma_inner)
                #params = xtol*(1-self.eps/2+self.eps*np.random.random(len(params),))
                
                #---Calculate the other fitness functions
                
                '''Add other fitness functions here
                first gotta calculate the spike trains again
                also look and see what the difference between
                the size of a pickle file and a text file is to 
                decide how to save stimulus and traces'''
                
                #Take the current best values and run the program again.
                #tempTimeIterStart=time.time()
                #(voltage_list, threshold_list, AScurrentMatrix_list, gridSpikeTime_list, interpolatedSpikeTime_list, \
                #    gridSpikeIndex_list, interpolatedSpikeVoltage_list, interpolatedSpikeThreshold_list) = \
                #    self.experiment.run_base_model(xtol)

                #timeFor1Iter=time.time()-tempTimeIterStart  

            # outer loop uses the outer standard deviation to randomize the initial values

            if stop_flag is True:
                break
            
            params = self.randomize_parameter_values(self.init_params, self.sigma_outer)


        # get the best one!
        min_error = float("inf")
        min_i = -1
        min_dt_multiplier = float("inf")
        
        for i, info in enumerate(self.iteration_info):
            if info['dt_multiplier'] < min_dt_multiplier:
                min_dt_multiplier = info['dt_multiplier']
        
        for i, info in enumerate(self.iteration_info):
            if info['error'] < min_error and info['dt_multiplier'] == min_dt_multiplier:
                min_error = info['error']
                min_i = i

        best_params = self.iteration_info[min_i]['out_params']

        self.experiment.set_neuron_parameters(best_params)

        logging.info('done optimizing')
        return best_params, self.init_params

    def run_once_bound(self, low_bound, high_bound):
        '''
        @param low_bound: a scalar initial guess for the optimizer
        @param high_bound: a scalar high bound for the optimizer
        @return: tuple including parameters that optimize function and value - see fmin docs
        '''                
        return fminbound(self.error_function, low_bound, high_bound, args=(self.experiment,self.error_function_data), maxfun=200, full_output=True )
        #Note is defined in the top level script
    
  
    def run_once(self, param0):
        '''
        @param param0: a list of the initial guesses for the optimizer
        @return: tuple including parameters that optimize function and value - see fmin docs
        '''       
#        fmin(func, x0, args=(), xtol=1e-4, ftol=1e-4, maxiter=None, maxfun=None, full_output=0, disp=1, retall=0, callback=None):

        print 'self.error_function_data', self.error_function_data
        xopt, fopt, _, _, _, _ = fmin(self.error_function, param0, args=(self.experiment,self.error_function_data),xtol=self.xtol, ftol=self.ftol,  maxiter=self.internal_iterations, maxfun=self.internal_iterations, retall=1,full_output=1, disp=1)

        return xopt, fopt
#         res = minimize(self.error_function, param0,
#                         method='Nelder-Mead',
#                         args=(self.experiment,self.error_function_data),  
#                         options={
#                                  'maxiter':self.internal_iterations, 
#                                  'xtol':self.xtol, 
#                                  'ftol':self.ftol,
#                                  'maxfun':self.internal_iterations, 
#                                  'retall':1,
#                                  'full_output':1, 
#                                  'disp':1}
#                        )

#         res = minimize(self.error_function, param0,
#                        jac=False,
#                        method='BFGS',
#                        args=(self.experiment,self.error_function_data),  
#                        options={
#                                 'maxiter':self.internal_iterations,
#                                 'epsilon':1e-8,
#                                 'gtol':1e-5,
#                                 'full_output':1, 
#                                 'disp':1}
#                        )
# 
#          
#         print res
#         return res.x, res.fun


#        #Note is defined in the top level script
#        def mycallback_ncg(xk):
#            print 'Using Newton-CG method, xk: ', xk
#        def mycallback_nm(xk):
#            print 'Using Nelder-Mead method, xk: ', xk
#        eps=1e-15
#        options={}
#  #      options['avextox']=eps
#        options['maxiter']=500
##        options['full_output']=True
##        options['disp']=True
##        options['retall']=True
#
#        print 'Using Newton-CG method'
#        iteration_start_time = time.time()
#        xopt = minimize(self.error_function, param0, args=(self.experiment,), method='Newton-CG', jac=f_prime_constructor(self.error_function), callback=mycallback_ncg, options=options, tol=eps)
#        print 'Newton-CG method took', (time.time()-iteration_start_time)/60., 'seconds'
#
##        print 'Using Nelder-Mead method'
##        iteration_start_time = time.time()
##        xopt = minimize(self.error_function, param0, args=(self.experiment,), method='Nelder-Mead', callback=mycallback_nm, options=options, tol=eps)
##        print 'Nelder-Mead method took', (time.time()-iteration_start_time)/60., 'seconds'
#
#        print xopt
#         return xopt, fopt



    



import os
import json
import logging
import scipy.ndimage.measurements as measurements
from scipy.spatial import cKDTree
from allensdk.brain_observatory.roi_masks import create_roi_mask
import pandas as pd
import numpy as np

CRITERIA_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             "resources",
                             "roi_filter_training_criteria.json")
_CRITERIA = None

def CRITERIA():
    global _CRITERIA
    if _CRITERIA is None:
        with open(CRITERIA_FILE, "r") as f:
            _CRITERIA = json.load(f)
    return _CRITERIA


class TrainingLabelClassifier(object):
    '''Very basic threshold_based classifier.

    Has a decision function that is just the number of distinct
    criteria met by the classifier. Criteria are defined as a list
    of strings used with pandas.DataFrame.eval.

    Parameters
    ----------
    criteria : list
        List of evaluation strings.
    '''
    def __init__(self, criteria):
        '''Constructor.'''
        if criteria is None:
            self.criteria = []
        else:
            self.criteria = criteria

    def decision_function(self, X):
        '''Get the distance from the decision boundary.

        Parameters
        ----------
        X : array-like
            Features for each ROI.

        Returns
        -------
        T : array-like
            Distance for each sample from the decision boundary.
        '''
        T = np.zeros((X.shape[0],), dtype=int)
        for crit in self.criteria:
            T[X.eval(crit).as_matrix()] += 1
        return T


class TrainingMultiLabelClassifier(object):
    '''Multilabel classifier using groups of TrainingLabelClassifiers.

    This was used to generate labeling for training the original SVM
    for classification.

    Parameters
    ----------
    criteria : dictionary
        Label names and criteria for each label.
    '''
    def __init__(self, criteria=None):
        '''Constructor.'''
        if criteria is None:
            criteria = CRITERIA()
        i = 0
        self.labels = sorted(criteria.keys())
        self._codes = {}
        self._classifiers = {}
        for label in self.labels:
            label_criteria = criteria[label]
            self._codes[label] = 2**i
            self._classifiers[2**i] = TrainingLabelClassifier(label_criteria)
            i += 1

    def _labels_as_columns(self, label_codes):
        '''Convert label series to boolean columns for each label.

        Parameters
        ----------
        label_codes : pandas.Series
            Label codes.

        Returns
        -------
        pandas.DataFrame
            Dataframe where each column is a label, and values are
            True for labeled or False otherwise.
        '''
        output = pd.DataFrame()
        for name in self.labels:
            number = self._codes[name]
            output[name] = (label_codes & number) > 0
        return output

    def _map_code_to_list(self, label_code):
        output = []
        for name in self.labels:
            number = self._codes[name]
            if (label_code & number) > 0:
                output.append(name)
        return output

    def get_eXcluded(self, X):
        '''Get the calculated value of the eXcluded column.

        This is useful for comparison with the original classifier
        implementation.

        Parameters
        ----------
        X : pandas.DataFrame
            Object features from the object list file.

        Returns
        -------
        numpy.ndarray
            Calculated eXcluded score from the classifier.
        '''
        eXcluded = np.zeros((X.shape[0],), dtype=X["eXcluded"].dtype)
        for classifier in self._classifiers.itervalues():
            eXcluded += classifier.decision_function(X)
        # match the object list values
        eXcluded[eXcluded > 0] += 10
        eXcluded[X["eXcluded"] == 1] = 1
        eXcluded[X["eXcluded"] == 2] = 2
        return eXcluded.as_matrix()

    def label_data(self, X, as_columns=True):
        '''Generate labels for each row in X.

        Parameters
        ----------
        X : pandas.DataFrame
            Object features from the object list file.

        Returns
        -------
        numpy.ndarray
            Array of label codes representing the combination of labels
            found for each row.
        '''
        labels = np.zeros((X.shape[0],), dtype=int)
        for label, classifier in self._classifiers.iteritems():
            labels[classifier.decision_function(X) > 0] += label
        if as_columns:
            return self._labels_as_columns(labels)
        else:
            return pd.Series(labels).apply(self._map_code_to_list)


def calculate_max_border(log0_df, max_shift):
    '''Calculate motion boundary from frame offsets.

    When the motion correction algorithm fails to find sufficient
    matches, it generates very large frame offsets. The use of
    `max_shift` avoids filtering too many cells due to the large
    offsets, with the tradeoff that those frames will be noise.

    Parameters
    ----------
    log0_df : pandas.DataFrame
        Dataframe containing the x, y offsets from motion correction.
    max_shift : float
        Maximum shift to allow when considering motion correction. Any
        larger shifts are considered outliers.

    Returns
    -------
    list
        [right_shift, left_shift, down_shift, up_shift]
    '''
    # strip outliers
    x_no_outliers = log0_df["x"][(log0_df["x"] >= -max_shift) & \
                                 (log0_df["x"] <= max_shift)]
    y_no_outliers = log0_df["y"][(log0_df["y"] >= -max_shift) & \
                                 (log0_df["y"] <= max_shift)]

    right_shift = np.max(-1*x_no_outliers.min(), 0)
    left_shift = np.max(x_no_outliers.max(), 0)
    down_shift = np.max(-1*y_no_outliers.min(), 0)
    up_shift = np.max(y_no_outliers.max(), 0)

    border = [right_shift, left_shift, down_shift, up_shift]

    if np.any(np.isnan(np.array(border))):
        raise ValueError("Motion correction failed.")

    return border


def order_rois_by_object_list(object_data, rois):
    '''Reorder rois by matching bounding boxes to object list.

    Parameters
    ----------
    object_data : pandas.DataFrame
        Object list data.
    rois : list
        List of RoiMasks.

    Returns
    -------
    list
        The list of rois reordered to index the same as object_data.
    '''
    object_points = object_data[["minx", "miny", "maxx", "maxy", "area"]].copy()
    object_points["maxx"] += 1
    object_points["maxy"] += 1
    roi_points = []
    for roi in rois:
        roi_points.append([roi.x, roi.y, roi.x+roi.width, roi.y+roi.height,
                           roi.mask.sum()])
    reorder_index = get_indices_by_distance(object_points,
                                            np.array(roi_points))
    multi_mapped = set()
    if len(set(reorder_index)) != reorder_index.shape[0]:
        unique, counts = np.unique(reorder_index, return_counts=True)
        multi_mapped = set(unique[counts > 1])
        not_mapped = set(np.setdiff1d(np.arange(reorder_index.shape[0]),
                                      reorder_index))
        logging.warning("ROIs don't uniquely map to object_list")
        for idx in (multi_mapped | not_mapped):
            logging.warning(
                "%s has ambiguous mapping to object list" % rois[idx].label)
    out_rois = []
    for i in reorder_index:
        roi = rois[i]
        if i in multi_mapped:
            roi.labels.append("duplicate")
        out_rois.append(roi)
    return out_rois


def get_rois(segmentation_stack, border=None):
    '''Extract a list of rois from the segmentation data array.

    Parameters
    ----------
    segmentation_stack : numpy.ndarray
        The array from the maxInt_masks file showing the object masks.
    border : list
        [right_shift, left_shift, down_shift, up_shift] bounding box
        determined from motion correction.

    Returns
    -------
    list
        List of RoiMask objects.
    '''
    rois = []
    if border is None:
        border = [0, 0, 0, 0]
    height = segmentation_stack.shape[1]
    width = segmentation_stack.shape[2]
    for i in range(segmentation_stack.shape[0]):
        page = segmentation_stack[i, :, :]
        label_mask, num_labels = measurements.label(
            page, structure=[[1, 1, 1], [1, 1, 1], [1, 1, 1]])
        for label in range(1, num_labels + 1):
            img_mask = label_mask == label
            mask = create_roi_mask(width, height, border,
                                   roi_mask=img_mask,
                                   label="ROI {}:{}".format(i, label),
                                   mask_group=i)
            mask.labels = []
            if mask.overlaps_motion_border:
                mask.labels.append("motion_border")
            rois.append(mask)
    return rois


def get_indices_by_distance(object_list_points, mask_points):
    '''Find indices of nearest neighbor matches.

    Require a distance of 0 (perfect match) and a unique match between
    masks and object_list entries.
    '''
    tree = cKDTree(mask_points)
    distance, indices = tree.query(object_list_points)
    if distance.max() > 0:
        logging.error("An ROI did not match object list exactly.")
        raise AssertionError("Max match distance greater than 0")
    return indices



import sys

import neuron_morphology.swc as swc
import neuron_morphology.features.feature_extractor as feature_extractor
from allensdk.internal.core.lims_pipeline_module import PipelineModule


########################################################################

def main(jin):
    try:
        swc_file = jin["swc_file"]
        xform = jin["pia_transform"]
        depth = jin["relative_soma_depth"]
    except:
        print("** Unable to find requisite fields in input json");
        raise


    ####################################################################
    # calculate features

    try:
        nrn = swc.read_swc(swc_file)
    except:
        print("** Error reading swc file")
        raise

    try:
        aff = []
        for i in range(12):
            aff.append(xform["tvr_%02d" % i])
        nrn.apply_affine(aff)
    except:
        print("** Error applying affine transform")
        raise

    #try:
    #    # save a copy of affine-corrected file
    #    tmp_swc_file = swc_file[:-4] + "_pia.swc"
    #    nrn.write(tmp_swc_file)
    #except:
    #    # treat this as a soft error and print a warning
    #    print("Note: unable to write copy of affine corrected pia file")

    try:
        features = feature_extractor.MorphologyFeatures(nrn, depth)
        data = {}
        data["axon"] = features.axon
        data["cloud"] = features.axon_cloud
        data["dendrite"] = features.dendrite
        data["basal_dendrite"] = features.basal_dendrite
        data["apical_dendrite"] = features.apical_dendrite
        data["all_neurites"] = features.all_neurites
    except:
        print("** Error calculating morphology features")
        raise

    # make output of new module backwards compatible with previous module
    md = {}
    feat = {}
    feat["number_of_stems"] = data["dendrite"]["num_stems"]
    feat["max_euclidean_distance"] = data["dendrite"]["max_euclidean_distance"]
    feat["max_path_distance"] = data["dendrite"]["max_path_distance"]
    feat["overall_depth"] = data["dendrite"]["depth"]
    feat["total_volume"] = data["dendrite"]["total_volume"]
    feat["average_parent_daughter_ratio"] = data["dendrite"]["mean_parent_daughter_ratio"]
    feat["average_diameter"] = data["dendrite"]["average_diameter"]
    feat["total_length"] = data["dendrite"]["total_length"]
    feat["nodes_over_branches"] = data["dendrite"]["neurites_over_branches"]
    feat["overall_width"] = data["dendrite"]["width"]
    feat["number_of_nodes"] = data["dendrite"]["num_nodes"]
    feat["average_bifurcation_angle_local"] = data["dendrite"]["bifurcation_angle_local"]
    feat["number_of_bifurcations"] = data["dendrite"]["num_bifurcations"]
    feat["average_fragmentation"] = data["dendrite"]["mean_fragmentation"]
    feat["number_of_tips"] = data["dendrite"]["num_tips"]
    feat["average_contraction"] = data["dendrite"]["contraction"]
    feat["average_bifuraction_angle_remote"] = data["dendrite"]["bifurcation_angle_remote"]
    feat["number_of_branches"] = data["dendrite"]["num_branches"]
    feat["total_surface"] = data["dendrite"]["total_surface"]
    feat["max_branch_order"] = data["dendrite"]["max_branch_order"]
    feat["soma_surface"] = data["dendrite"]["soma_surface"]
    feat["overall_height"] = data["dendrite"]["height"]


    md["features"] = feat
    data["morphology_data"] = md

    return data


if __name__=='__main__':
    module = PipelineModule()
    jin = module.input_data()
    jout = main(jin)
    module.write_output_data(jout)



import h5py
import math

class LabNotebookReader(object):
    def __init__(self):
        self.register_enabled_names()

    # mapping of notebook keys to keys representing if that value is
    #   enabled
    # move this to subclasses if/when key names diverge
    def register_enabled_names(self):
        self.enabled = {}
        self.enabled["V-Clamp Holding Level"] = "V-Clamp Holding Enable"
        self.enabled["RsComp Bandwidth"] = "RsComp Enable"
        self.enabled["RsComp Correction"] = "RsComp Enable"
        self.enabled["RsComp Prediction"] = "RsComp Enable"
        self.enabled["Whole Cell Comp Cap"] = "Whole Cell Comp Enable"
        self.enabled["Whole Cell Comp Resist"] = "Whole Cell Comp Enable"
        self.enabled["I-Clamp Holding Level"] = "I-Clamp Holding Enable"
        self.enabled["Neut Cap Value"] = "Neut Cap Enable"
        self.enabled["Bridge Bal Value"] = "Bridge Bal Enable"


    # lab notebook has two sections, one for numeric data and the other
    #   for text data. this is an internal function to fetch data from
    #   the numeric part of the notebook
    def get_numeric_value(self, name, data_col, sweep_col, enable_col, sweep_num, default_val):
        data = self.val_number
        # val_number has 3 dimensions -- the first has a shape of
        #   (#fields * 9). there are many hundreds of elements in this
        #   dimension. they look to represent the full array of values
        #   (for each field for each multipatch) for a given point in
        #   time, and thus given sweep
        # according to Thomas Braun (igor nwb dev), the first 8 pages are
        #   for headstage data, and the 9th is for headstage-independent
        #   data
        # return value is last non-empty entry in specified column
        #   for specified sweep number
        return_val = default_val
        for sample in data:
            swp = sample[sweep_col][0]
            if math.isnan(swp):
                continue
            if int(swp) == sweep_num:
                if enable_col is not None and sample[enable_col][0] != 1.0:
                    continue # 'enable' flag present and it's turned off
                val = sample[data_col][0]
                if not math.isnan(val):
                    return_val = val
        return return_val

    # internal function for fetching data from the text part of the notebook
    def get_text_value(self, name, data_col, sweep_col, enable_col, sweep_num, default_val):
        data = self.val_text
        # algorithm mirrors get_numeric_value
        # return value is last non-empty entry in specified column
        #   for specified sweep number
        return_val = default_val
        for sample in data:
            swp = sample[sweep_col][0]
            if len(swp) == 0:
                continue
            if int(swp) == int(sweep_num):
                if enable_col is not None: # and sample[enable_col][0] != 1.0:
                    # this shouldn't happen, but if it does then bitch
                    #   as this situation hasn't been tested (eg, is 
                    #   enabled indicated by 1.0, or "1.0" or "true" or ??)
                    Exception("Enable flag not expected for text values")
                    #continue # 'enable' flag present and it's turned off
                val = sample[data_col][0]
                if len(val) > 0:
                    return_val = val
        return return_val

    # looks for key in lab notebook and returns the value associated with
    #   the specified sweep, or the default value if no value is found
    #   (NaN and empty strings are considered to be non-values)
    def get_value(self, name, sweep_num, default_val):
        # name_number has 3 dimensions -- the first has shape
        #   (#fields * 9) and stores the key names. the second looks
        #   to store units for those keys. The third is numeric text
        #   but it's role isn't clear
        numeric_fields = self.colname_number[0]
        text_fields = self.colname_text[0]
        # val_number has 3 dimensions -- the first has a shape of
        #   (#fields * 9). there are many hundreds of elements in this
        #   dimension. they look to represent the full array of values
        #   (for each field for each multipatch) for a given point in
        #   time, and thus given sweep
        if name in numeric_fields:
            sweep_idx = numeric_fields.tolist().index("SweepNum")
            enable_idx = None
            if name in self.enabled:
                enable_col = self.enabled[name]
                enable_idx = numeric_fields.tolist().index(enable_col)
            field_idx = numeric_fields.tolist().index(name)
            return self.get_numeric_value(name, field_idx, sweep_idx, enable_idx, sweep_num, default_val)
        elif name in text_fields:
            # first check to see if file includes old version of column name
            if "Sweep #" in text_fields:
                sweep_idx = text_fields.tolist().index("Sweep #")
            else:
                sweep_idx = text_fields.tolist().index("SweepNum")
            enable_idx = None
            if name in self.enabled:
                enable_col = self.enabled[name]
                enable_idx = text_fields.tolist().index(enable_col)
            field_idx = text_fields.tolist().index(name)
            return self.get_text_value(name, field_idx, sweep_idx, enable_idx, sweep_num, default_val)
        else:
            return default_val
            
        

""" Loads lab notebook data out of a first-generation IVSCC NWB file,
    that was manually translated from the IGOR h5 dump.
    Notebook data can be read through get_value() function
"""
class LabNotebookReaderIvscc(LabNotebookReader):
    def __init__(self, nwb_file, h5_file):
        LabNotebookReader.__init__(self)
        # for lab notebook, select first group 
        h5 = h5py.File(h5_file, "r")
        #
        # TODO FIXME check notebook version... but how?
        #
        notebook = h5["MIES/LabNoteBook/ITC18USB/Device0"]
        # load column data into memory
        self.colname_number = notebook["KeyWave/keyWave"].value
        self.val_number = notebook["settingsHistory/settingsHistory"].value
        self.colname_text = notebook["TextDocKeyWave/txtDocKeyWave"].value
        self.val_text = notebook["textDocumentation/txtDocWave"].value
        h5.close()



########################################################################
########################################################################
""" Loads lab notebook data out of an Igor-generated NWB file.
    Module input is the name of the nwb file.
    Notebook data can be read through get_value() function
"""
class LabNotebookReaderIgorNwb(LabNotebookReader):
    def __init__(self, nwb_file):
        LabNotebookReader.__init__(self)
        # for lab notebook, select first group 
        # NOTE this probably won't work for multipatch
        h5 = h5py.File(nwb_file, "r")
        #
        # TODO FIXME check notebook version
        #
        for k in h5["general/labnotebook"]:
            notebook = h5["general/labnotebook"][k]
            break
        # load column data into memory
        self.val_text = notebook["textualValues"].value
        self.colname_text = notebook["textualKeys"].value
        self.val_number = notebook["numericalValues"].value
        self.colname_number = notebook["numericalKeys"].value
        h5.close()
        #
        self.register_enabled_names()

        

# creates LabNotebookReader appropriate to ivscc-NWB file version
def create_lab_notebook_reader(nwb_file, h5_file=None):
    pass
    h5 = h5py.File(nwb_file, "r")
    if "general/labnotebook" in h5:
        version = "IgorNwb"
    else:
        version = "IgorH5"
    h5.close()
    if version == "IgorNwb":
        return LabNotebookReaderIgorNwb(nwb_file)
    elif version == "IgorH5":
        return LabNotebookReaderIvscc(nwb_file, h5_file)
    else:
        Exception("Unable to determine NWB input type")




# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import copy
import warnings
import functools
from numpy import VisibleDeprecationWarning

    
def deprecated(message=None):

    if message is None:
        message = '' 
    
    def output_decorator(fn):
        
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
        
            warnings.warn("Function {0} is deprecated. {1}".format(
                          fn.__name__, message), 
                          category=VisibleDeprecationWarning, stacklevel=2)
            
            return fn(*args, **kwargs)
            
        return wrapper
        
    return output_decorator
    
    
def class_deprecated(message=None):

    if message is None:
        message = ''
        
    def output_class_decorator(cls):
        
        fn_copy = copy.deepcopy(cls.__init__)
        
        @functools.wraps(cls.__init__)
        def wrapper(*args, **kwargs):
            warnings.warn("Class {0} is deprecated. {1}".format(
                          cls.__name__, message), 
                          category=VisibleDeprecationWarning, stacklevel=2)
            fn_copy(*args, **kwargs)
                          
        cls.__init__ = wrapper
        return cls
        
    return output_class_decorator



'''Written by Corinne Teeter 3-31-14
'''

import matplotlib
import matplotlib.pyplot as plt
import numpy as np


def checkPreprocess(originalStim_list, processedStim_list, originalVoltage_list, processedVoltage_list, config, blockME=False):
        
    timeOriginal=np.arange(len(np.concatenate(originalStim_list)))*config.neuron['dt']
    if 'subSample' in config.dictOfPreprocessMethods.keys():
        timeProcessed=np.arange(len(np.concatenate(processedStim_list)))*config.dictOfPreprocessMethods['subSample']['desired_time_step']
    else:
        timeProcessed=timeOriginal
        
    plt.figure(figsize=(20,10))
    plt.subplot(4,1,1)
    plt.plot(timeOriginal, np.concatenate(originalStim_list), 'b')
    plt.title('original stimulation')
    
    plt.subplot(4,1,2)
    plt.plot(timeProcessed, np.concatenate(processedStim_list), 'r')
    plt.title('processed stimulation')        

    plt.subplot(4,1,3)
    plt.plot(timeOriginal, np.concatenate(originalVoltage_list), 'b')
    plt.title('original voltage')
    
    plt.subplot(4,1,4)
    plt.plot(timeProcessed, np.concatenate(processedVoltage_list), 'r')
    plt.title('processed voltage') 
    
    plt.annotate(config.cellName+': View result of preprocessing', xy=(.4, .975),
         xycoords='figure fraction',
         horizontalalignment='left', verticalalignment='top',
         fontsize=20)

#    plt.show(block=blockME)

def plotSpikes(voltage_list, spike_ind_list, dt, blockME=False, method=False):
    
    converted_spike_ind_list=[]
    time=np.arange(len(np.concatenate(voltage_list)))*dt
    #--find the length of each vector
    thelength=0
    for ii, voltage in enumerate(voltage_list):
        converted_spike_ind_list.append(spike_ind_list[ii]+thelength)
        thelength=thelength+len(voltage)
    
    subsampled_time=[time[ii] for ii in np.concatenate(converted_spike_ind_list)]

    plt.figure(figsize=(20, 5))
    plt.plot(time, np.concatenate(voltage_list), 'b')
    plt.plot(subsampled_time, [np.concatenate(voltage_list)[ii] for ii in np.concatenate(converted_spike_ind_list)], 'r.', ms=16)
    if method==False:
        plt.title('Spikes')
    else:
        plt.title('Spikes. Method used: '+method)
    
    plt.ylabel('voltage (V)')
    plt.xlabel('time (s)')
#    plt.show(block=blockME)
    
def checkSpikeCutting(originalStim_list, cutStim_list, originalVoltage_list, cutVoltage_list, allindOfNonSpiking_list, config, blockME=False):
    
    if len(originalStim_list)!=len(cutStim_list) or  \
        len(originalStim_list)!=len(originalVoltage_list) or \
        len(originalStim_list)!=len(cutVoltage_list) or \
        len(originalStim_list)!=len(allindOfNonSpiking_list):
        raise Exception('lists are not the same length')
    
        
    lengthGoingToAdd=0
    whole_ind=np.array([])
    whole_v=np.array([])
    for trace, ind_array in zip(originalVoltage_list, allindOfNonSpiking_list):
        ind=np.arange(0, len(trace))+lengthGoingToAdd
        whole_ind=np.append(whole_ind, [ind[ii] for ii in ind_array])
        whole_v=np.append(whole_v, [trace[ii] for ii in ind_array])
        lengthGoingToAdd=lengthGoingToAdd+len(trace)
             
             
    time=np.arange(len(np.concatenate(originalStim_list)))*config.neuron['dt']        
    plt.figure(figsize=(20,10))
    
    plt.subplot(1,1,1)
    plt.plot(time, np.concatenate(originalVoltage_list))
    plt.title('voltage')
    
    plt.plot(whole_ind*config.neuron['dt'], whole_v, '--r', lw=2)
    
    plt.annotate(config.cellName+': check spike cutting', xy=(.4, .975),
                 xycoords='figure fraction',
                 horizontalalignment='left', verticalalignment='top',
                 fontsize=20)
#    plt.show(block=blockME)

def plotLineRegress1(slope, intercept, r,xlim):
    y=slope*xlim+intercept
    print 'slope=', slope, 'intercept=', intercept, 'xlim', xlim
    plt.plot(xlim, y, '-k', lw=4, label='slope='+"%.2f"%slope+', intercept='+"%.3f"%intercept+', r='+"%.2f"%r)

def plotLineRegressRed(slope, intercept, r,xlim):
    y=slope*xlim+intercept
    print 'slope=', slope, 'intercept=', intercept, 'xlim', xlim
    plt.plot(xlim, y, '-r', lw=4, label='slope='+"%.2f"%slope+', intercept='+"%.3f"%intercept+', r='+"%.2f"%r)



import numpy as np
import logging

from allensdk.internal.model.glif.find_spikes import find_spikes_list

from allensdk.ephys.extract_cell_features import get_stim_characteristics

def least_squares_RCEl_calc_tested(voltage_list, current_list, dt):
    '''Calculate resistance, capacitance and resting potential by performing 
    least squares on current and voltage.
    
    Parameters
    ----------
    voltage_list: list of arrays 
        voltage responses for several sweep repeats
    current_list: list of arrays 
        current injections for several sweep repeats
    dt: float
        time step size in voltage and current traces
    
    Returns
    -------
    r_list: list of floats
        each value corresponds to the resistance of a sweep
    c_list: list of floats 
        each value corresponds to the capacitance of a sweep
    el_list: list of floats 
        each value corresponds to the resting potential of a sweep
        '''    

    r_list=[]
    c_list=[]
    el_list=[]
    for voltage, current in zip(voltage_list, current_list):
        matrix=np.ones((len(voltage)-1, 3))
        matrix[:,0]=voltage[0:len(voltage)-1]
        matrix[:,1]=current[0:len(current)-1]
        lsq_non_der_raw=np.linalg.lstsq(matrix, voltage[1:])[0]
#        (r_lsq_non_der_raw, c_lsq_non_der_raw, El_lsq_non_der_raw)=RCEL_from_standard_space(lsq_non_der_raw) 
        
        c_lsq_non_der_raw=dt/lsq_non_der_raw[1]
        r_lsq_non_der_raw=-lsq_non_der_raw[1]/(lsq_non_der_raw[0]-1)
        El_lsq_non_der_raw=-lsq_non_der_raw[2]/(lsq_non_der_raw[0]-1)   
        r_list.append(r_lsq_non_der_raw)
        c_list.append(c_lsq_non_der_raw)
        el_list.append(El_lsq_non_der_raw)
    
    return r_list, c_list, el_list





import logging
import numpy as np

# TODO: license
# TODO: document

class GlifExperiment( object ):
    def __init__(self, neuron, dt, stim_list, resp_list,
                 spike_time_steps, grid_spike_times, grid_spike_voltages,
                 param_fit_names,
                 **kwargs):

        self.neuron = neuron
        self.dt = dt
        self.stim_list = stim_list
        self.resp_list = resp_list
        self.spike_time_steps = spike_time_steps
        self.grid_spike_times = grid_spike_times
        self.grid_spike_voltages = grid_spike_voltages
        self.param_fit_names = param_fit_names

        self.spike_errors = []

        
    def run(self, param_guess):
        '''This code will run the loaded neuron model in reference to the target neuron spikes.
        inputs:
            self: is the instance of the neuron model and parameters alone with the values of the target spikes.
                NOTE the values in each array of the self.gridSpikeIndexTarge_list and the self.interpolated_spike_times
                are in reference to the time start of of the stim in each induvidual array (not the universal time)
            param_guess: array of scalars of the values that will be inserted into the mapping function below.
        returns:
            voltage_list: list of array of voltage values. NOTE: IF THE MODEL NEURON SPIKES BEFORE THE TARGET THE VOLTAGE WILL 
                NOT BE CALCULATED THEREFORE THE RESULTING VECTOR WILL NOT BE AS LONG AS THE TARGET AND ALSO WILL NOT 
                MAKE SENSE WITH THE STIMULUS UNLESS YOU CUT IT AND OUTPUT IT TOO.
            grid_spike_times_list:
            interpolated_spike_time_list: an array of the actual times of the spikes. NOTE: THESE TIMES ARE CALCULATED BY ADDING THE 
                TIME OF THE INDIVIDUAL SPIKE TO THE TIME OF THE LAST SPIKE.
            gridISIFromLastTargSpike_list: list of arrays of spike times of the model in reference to the last target (biological) 
                spike (not in reference to sweep start)
            interpolatedISIFromLastTargSpike_list: list of arrays of spike times of the model in reference to the last target (biological) 
                spike (not in reference to sweep start)
            voltageOfModelAtGridBioSpike_list: list of arrays of scalars that contain the voltage of the model neuron when the target or bio neuron spikes.    
            theshOfModelAtGridBioSpike_list: list of arrays of scalars that contain the threshold of the model neuron when the target or bio neuron spikes.'''

        self.set_neuron_parameters(param_guess)    
        self.spike_errors = []

        run_data = []
        
        for stim_list_index in xrange(len(self.stim_list)):
            run_data.append(self.neuron.run_with_biological_spikes(self.stim_list[stim_list_index],
                                                                   self.resp_list[stim_list_index],
                                                                   self.spike_time_steps[stim_list_index]))
          
        return {
            'voltage': [ rd['voltage'] for rd in run_data ],
            'threshold': [ rd['threshold'] for rd in run_data ],
            'AScurrent_matrix': [ rd['AScurrent_matrix'] for rd in run_data ],

            'grid_ISI': [ rd['grid_ISI'] for rd in run_data ],
            'interpolated_ISI': [ rd['interpolated_ISI'] for rd in run_data ],

            'grid_model_spike_times': [ rd['grid_model_spike_times'] for rd in run_data ],
            'interpolated_model_spike_times': [ rd['interpolated_model_spike_times'] for rd in run_data ],
            
            'grid_model_spike_voltages': [ rd['grid_model_spike_voltages'] for rd in run_data ],
            'interpolated_model_spike_voltages': [ rd['interpolated_model_spike_voltages'] for rd in run_data ],

            'grid_bio_spike_model_voltage': [ rd['grid_bio_spike_model_voltage'] for rd in run_data ],
            'grid_bio_spike_model_threshold': [ rd['grid_bio_spike_model_threshold'] for rd in run_data ]
        }


    def run_base_model(self, param_guess):
        '''This code will run the loaded neuron model.
        inputs:
            self: is the instance of the neuron model and parameters alone with the values of the target spikes.
                NOTE the values in each array of the self.gridSpikeIndexTarge_list and the self.interpolated_spike_times
                are in reference to the time start of of the stim in each induvidual array (not the universal time)
            param_guess: array of scalars of the values that will be inserted into the mapping function below.
        returns:
            voltage_list: list of array of voltage values. NOTE: IF THE MODEL NEURON SPIKES BEFORE THE TARGET THE VOLTAGE WILL 
                NOT BE CALCULATED THEREFORE THE RESULTING VECTOR WILL NOT BE AS LONG AS THE TARGET AND ALSO WILL NOT 
                MAKE SENSE WITH THE STIMULUS UNLESS YOU CUT IT AND OUTPUT IT TOO.
            gridTime_list:
            interpolatedTime_list: an array of the actual times of the spikes. NOTE: THESE TIMES ARE CALCULATED BY ADDING THE 
                TIME OF THE INDIVIDUAL SPIKE TO THE TIME OF THE LAST SPIKE.
            grid_ISI_list: list of arrays of spike times of the model in reference to the last target (biological) 
                spike (not in reference to sweep start)
            interpolated_ISI_list: list of arrays of spike times of the model in reference to the last target (biological) 
                spike (not in reference to sweep start)
            grid_spike_voltage_list: list of arrays of scalars that contain the voltage of the model neuron when the target or bio neuron spikes.    
            grid_spike_threshold_list: list of arrays of scalars that contain the threshold of the model neuron when the target or bio neuron spikes.'''

        
        stim_list = self.stim_list
        
        self.set_neuron_parameters(param_guess)    
        self.spike_errors = []
        
        run_data = []

        for stim_list_index in xrange(len(stim_list)):
            run_data.append(self.neuron.run(stim_list[stim_list_index]))

          
        return {
            'voltage': [ rd['voltage'] for rd in run_data ],
            'threshold': [ rd['threshold'] for rd in run_data ],
            'AScurrents': [ rd['AScurrents'] for rd in run_data ],

            'spike_time_steps': [ rd['spike_time_steps'] for rd in run_data ],
            'grid_spike_times': [ rd['grid_spike_times'] for rd in run_data ],
            'interpolated_spike_times': [ rd['interpolated_spike_times'] for rd in run_data ],

            'interpolated_spike_voltage': [ rd['interpolated_spike_voltage'] for rd in run_data ],
            'interpolated_spike_threshold': [ rd['interpolated_spike_threshold'] for rd in run_data ]
        }

    def neuron_parameter_count(self):
        count = 0
        for fit_name in self.param_fit_names:
            try:
                coeff = self.neuron.coeffs[fit_name]
            except KeyError:
                logging.error("Neuron coefficient %s does not exist" % fit_name)
                raise

            # is it a list?
            try:
                # this will throw a type error if 'coeff' is a scalar
                coeff_size = len(coeff)
                count += coeff_size
            except TypeError:
                count += 1
        return count

    def set_neuron_parameters(self, param_guess): 
        '''Maps the parameter guesses to the coefficients of the model.  
        input:
            param_guess is vector of values.  It is assumed that the length will be '''

        index = 0
        for fit_name in self.param_fit_names:
            try:
                coeff = self.neuron.coeffs[fit_name]
            except KeyError:
                logging.error("Neuron coefficient %s does not exist" % fit_name)
                raise

            # is it a list?
            try:
                # this will throw a type error if 'coeff' is a scalar
                coeff_size = len(coeff)
                self.neuron.coeffs[fit_name] = param_guess[index:index+coeff_size]
                index += coeff_size
            except TypeError:
                self.neuron.coeffs[fit_name] = param_guess[index]
                index += 1




#!/usr/bin/python
import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt
import logging
import numpy as np
from allensdk.brain_observatory.r_neuropil import estimate_contamination_ratios
import allensdk.internal.core.lims_utilities as lu
import h5py
import json
import copy
import os
import sys
import argparse
import shutil

from allensdk.internal.core.lims_pipeline_module import PipelineModule, run_module

def debug(experiment_id, local=False):
    OUTPUT_DIRECTORY = "/data/informatics/CAM/neuropil"
    SDK_PATH = "/data/informatics/CAM/neuropil/allensdk"
    SCRIPT = "/data/informatics/CAM/neuropil/allensdk/allensdk/internal/pipeline_modules/run_neuropil_correction.py"

    exp = lu.query("select * from ophys_experiments where id = %d" % experiment_id)[0]
    sd = exp["storage_directory"]

    test_file = "/data/informatics/CAM/demix/%d/demixed_traces.h5" % experiment_id
    if os.path.exists(test_file):
        roi_trace_file = test_file
    else:
        roi_trace_file = os.path.join(sd, "demix", "%d_demixed_traces.h5" % experiment_id)

    exp_dir = os.path.join(OUTPUT_DIRECTORY, str(experiment_id))
    input_data = dict(
        roi_trace_file = roi_trace_file,
        neuropil_trace_file = os.path.join(sd, "processed", "neuropil_traces.h5"),
        storage_directory = exp_dir
        )

    run_module(SCRIPT, 
               input_data, 
               exp_dir,
               sdk_path=SDK_PATH,
               pbs=dict(vmem=160,
                        job_name="np_%d"% experiment_id,
                        walltime="36:00:00"),
               local=local,
               optional_args=['--log-level','DEBUG'])
    

def debug_plot(file_name, roi_trace, neuropil_trace, corrected_trace, r, r_vals=None, err_vals=None):
    fig = plt.figure(figsize=(15,10))

    ax = fig.add_subplot(211)
    ax.plot(roi_trace,'r', label="raw")
    ax.plot(corrected_trace,'b', label="fc")
    ax.plot(neuropil_trace,'g', label="neuropil")
    ax.set_xlim(0,roi_trace.size)
    ax.set_title('raw(%.02f, %.02f) fc(%.02f, %.02f) r(%f)' % (roi_trace.min(), roi_trace.max(), corrected_trace.min(), corrected_trace.max(), r))
    ax.legend()
    
    if r_vals is not None:
        ax = fig.add_subplot(212)
        ax.plot(r_vals, err_vals, "o")

    plt.savefig(file_name)
    plt.close()

def adjust_r_for_negativity(r, F_C, F_M, F_N):
    # this function is no longer used, but leaving it here just in case
    # loop through all of the negative spots and pick r to fix them
    neg_is = np.argwhere(F_C < 0)
    if neg_is.size > 0:
        logging.debug("Correcting for negative trace, starting with r = %f", r)

        for i in neg_is:
            if F_C[i] >= 0:
                continue
            # r_new = (F_C[i] + r_old * F_N[i]) / F_N[i]
            r = F_M[i] / F_N[i]
            F_C = F_M - r * F_N
            logging.debug("  updated r to %f", r)


        # if there is still a negative spot, it's off by some tiny epsilon.
        # step r down by delta_r increments until we find one that works.
        delta_r = -1e-5
        while F_C.min() < 0 and r >= 0.0:
            r += delta_r
            F_C = F_M - r * F_N
            logging.debug("  stepped r to %f", r)

        logging.debug("  finished with r = %f", r)

    return r


def main():
    module = PipelineModule()
    args = module.args

    jin = module.input_data()

    ########################################################################
    # prelude -- get processing metadata

    trace_file = jin["roi_trace_file"]
    neuropil_file = jin["neuropil_trace_file"]
    storage_dir = jin["storage_directory"]

    plot_dir = os.path.join(storage_dir, "neuropil_subtraction_plots")
    if os.path.exists(plot_dir):
        shutil.rmtree(plot_dir)

    try:
        os.makedirs(plot_dir)
    except:
        pass

    logging.info("Neuropil correcting '%s'", trace_file)

    ########################################################################
    # process data

    try:
        roi_traces = h5py.File(trace_file, "r")
    except:
        logging.error("Error: unable to open ROI trace file '%s'", trace_file)
        raise

    try:
        neuropil_traces = h5py.File(neuropil_file, "r")
    except:
        logging.error("Error: unable to open neuropil trace file '%s'", neuropil_file)
        raise

    '''
    get number of traces, length, etc.
    '''
    num_traces, T = roi_traces['data'].shape
    T_orig = T
    T_cross_val = int(T/2)
    if (T - T_cross_val > T_cross_val):
        T = T - 1

    # make sure that ROI and neuropil trace files are organized the same
    n_id = neuropil_traces["roi_names"]
    r_id = roi_traces["roi_names"]
    logging.info("Processing %d traces", len(n_id))
    assert len(n_id) == len(r_id), "Input trace files are not aligned (ROI count)"
    for i in range(len(n_id)):
        assert n_id[i] == r_id[i], "Input trace files are not aligned (ROI IDs)"
    '''
    initialize storage variables and analysis routine
    '''
    r_list = [ None ] * num_traces
    RMSE_list = [ -1 ] * num_traces
    roi_names = n_id
    corrected = np.zeros((num_traces, T_orig))
    r_vals = [ None ] * num_traces

    for n in range(num_traces):
        roi = roi_traces['data'][n]
        neuropil = neuropil_traces['data'][n]   

        if np.any(np.isnan(neuropil)):
            logging.warning("neuropil trace for roi %d contains NaNs, skipping", n)
            continue
            
        if np.any(np.isnan(roi)):
            logging.warning("roi trace for roi %d contains NaNs, skipping", n)
            continue

        r = None

        logging.info("Correcting trace %d (roi %s)", n, str(n_id[n]))
        results  = estimate_contamination_ratios(roi, neuropil)
        logging.info("r=%f err=%f it=%d", results["r"], results["err"], results["it"])

        r = results["r"]
        fc = roi - r * neuropil
        RMSE_list[n] = results["err"]
        r_vals[n] = results["r_vals"]
        
        debug_plot(os.path.join(plot_dir, "initial_%04d.png" % n),
                   roi, neuropil, fc, r, results["r_vals"], results["err_vals"])

        # mean of the corrected trace must be positive
        if fc.mean() > 0:
            r_list[n] = r
            corrected[n,:] = fc
        else:
            logging.warning("fc has negative baseline, skipping this r value")

    # compute mean valid r value 
    r_mean = np.array([r for r in r_list if r is not None ]).mean()

    # fill in empty r values
    for n in range(num_traces):        
        roi = roi_traces['data'][n]
        neuropil = neuropil_traces['data'][n]

        if r_list[n] is None:
            logging.warning("Error estimated r for trace %d. Setting to zero.", n)
            r_list[n] = 0
            corrected[n,:] = roi

        # save a debug plot
        debug_plot(os.path.join(plot_dir, "final_%04d.png" % n),
                   roi, neuropil, corrected[n,:], r_list[n])

        # one last sanity check
        eps = -0.0001
        if np.mean(corrected[n,:]) < eps:
            raise Exception("Trace %d baseline is still negative value after correction" % n)

        if r_list[n] < 0.0:
            raise Exception("Trace %d ended with negative r" % n)
        

    ########################################################################
    # write out processed data

    try:
        savefile = os.path.join(storage_dir, "neuropil_correction.h5")
        hf = h5py.File(savefile, 'w')
        hf.create_dataset("r", data=r_list)
        hf.create_dataset("RMSE", data=RMSE_list)
        hf.create_dataset("FC", data=corrected, compression="gzip")
        hf.create_dataset("roi_names", data=roi_names)

        for n in range(num_traces):
            r = r_vals[n]
            if r is not None:
                hf.create_dataset("r_vals/%d" % n, data=r)
        hf.close()
    except:
        logging.error("Error creating output h5 file")
        raise   
    
    roi_traces.close()
    neuropil_traces.close()

    jout = copy.copy(jin)
    jout["neuropil_correction"] = savefile
    module.write_output_data(jout)

    logging.info("finished")

if __name__ == "__main__": main()



#!/usr/bin/python
import sys
import psycopg2
import psycopg2.extras
sys.path.append("/home/keithg/allen/allensd")
import allensdk.core.json_utilities as json


def prep_json(spec_id):
    jin = {}

    try:
        conn_string = "host='limsdb2' dbname='lims2' user='atlasreader' password='atlasro'"
        conn = psycopg2.connect(conn_string)
        cursor = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
    except:
        print "unable to connect"
        raise

    ####################################################################
    # get polygons outlining layers in 20x image
    layer_sql = """
        select st.acronym, poly.path, wkf.storage_directory, wkf.filename, ims.id from image_series ims
        join sub_images si on si.image_series_id = ims.id
        join avg_graphic_objects layer on layer.sub_image_id = si.id
        join avg_graphic_objects poly on poly.parent_id = layer.id
        join avg_group_labels layert on layert.id = layer.group_label_id
        left join structures st on st.id = poly.structure_id
        join specimens hemisl on hemisl.id = ims.specimen_id
        join specimens cell on cell.parent_id = hemisl.id
        join neuron_reconstructions nr on nr.specimen_id = cell.id
        join well_known_files wkf on wkf.attachable_id = nr.id
        JOIN well_known_file_types wkft on wkft.id = wkf.well_known_file_type_id
        where nr.superseded is false
        and layert.name = 'Default'
        AND wkft.name = '3DNeuronReconstruction'
        and cell.id = %d 
        order by 1
    """
    cursor.execute(layer_sql % spec_id)
    #print layer_sql % spec_id
    poly_info = cursor.fetchall()
    if len(poly_info) == 0:
        print("Error -- cannot no polygon data for Specimen %d" % spec_id)
        sys.exit(1)
    poly = []
    for entry in poly_info:
        label = entry[0]
        path = entry[1]
        block = {}
        block["path"] = path
        block["label"] = label
        poly.append(block)
        ## break down string path into two numeric arrays
        #path_array = np.array(path.split(','))
        #path_x = np.array(path_array[0::2], dtype=float)
        #path_y = np.array(path_array[1::2], dtype=float)
        #block["path_array"] = path_array
        #block["path_x"] = path_x
        #block["path_y"] = path_y
        #poly[label] = block
    jin["layers"] = poly
    jin["storage_directory"] = poly_info[0][2]
    jin["swc_file"] = poly_info[0][3]

    # reconstruction ID
    # steal this from file name
    fname = jin["swc_file"].split("_m.swc")[0]
    reconstruction_id = int(fname[-9:])
    jin["reconstruction_id"] = reconstruction_id
    jin["resolution"] = 0.363


    # it appears that we need to restrict image query to use this ims_id
    ims_id = poly_info[0][4]

    ####################################################################
    # get soma outline
    soma_sql = """
        SELECT poly.path
        from image_series ims
        join sub_images si on si.image_series_id = ims.id 
        join avg_graphic_objects layer on layer.sub_image_id = si.id
        join avg_graphic_objects poly on poly.parent_id = layer.id
        join avg_group_labels layert on layert.id = layer.group_label_id
        left JOIN images im ON im.id=si.image_id
        left JOIN scans sc ON sc.image_id=im.id
        JOIN avg_group_labels agl ON layer.group_label_id=agl.id
        left join structures st on st.id = poly.structure_id
        join specimens hemisl on hemisl.id = ims.specimen_id
        join specimens cell on cell.parent_id = hemisl.id
        join neuron_reconstructions nr on nr.specimen_id = cell.id
        join well_known_files wkf on wkf.attachable_id = nr.id
        JOIN well_known_file_types wkft ON wkft.id = wkf.well_known_file_type_id AND wkft.name = '3DNeuronReconstruction'
        where nr.superseded is false
        and agl.name = 'Soma'
        and cell.id = %d
    """
    #print soma_sql % spec_id
    cursor.execute(soma_sql % spec_id)
    soma_res = cursor.fetchall()
    som = {}
    som["label"] = "Soma"
    som["path"] = soma_res[0]
    jin["soma"] = som

    ####################################################################
    # get 20x image
    img_sql = """
        SELECT ss.storage_directory, im.jp2 from image_series ims
        join sub_images si on si.image_series_id = ims.id 
        left JOIN images im ON im.id=si.image_id
        left JOIN specimens cell on ims.specimen_id = cell.id
        join slides ss on ss.id = im.slide_id
        where cell.id = %d
    """

    # get 20x image
    img_sql = """
        SELECT ss.storage_directory, im.jp2
        from image_series ims
        join sub_images si on si.image_series_id = ims.id 
        left JOIN images im ON im.id=si.image_id
        join slides ss on ss.id = im.slide_id
        and ims.id = %s
    """
    try:
        cursor.execute(img_sql % ims_id)
        #cursor.execute(img_sql % spec_id)
        img_res = cursor.fetchall()
        img_path = img_res[0][0] + img_res[0][1]
        #img_path = "%s-20x.jpeg" % str(spec_id)
    except:
        print "Error fetching path to 20x image from database"
        print img_sql % spec_id
        raise

    img = {}
    #img["img_res"] = img_res
    img["img_path"] = img_path
    jin["20x"] = img

    return jin

if __name__ == "__main__":
    spec_id = 490387590
    jin = prep_json(spec_id)
    print("Test mode: creating input.json for specimen.id=%d" % spec_id)
    json.write("input.json", jin)




import logging
import cPickle
import allensdk.internal.core.lims_utilities as lu
from allensdk.internal.core.lims_pipeline_module import PipelineModule, run_module
from allensdk.internal.brain_observatory import roi_filter, roi_filter_utils
from allensdk.brain_observatory.roi_masks import (RIGHT_SHIFT, LEFT_SHIFT,
                                                  DOWN_SHIFT, UP_SHIFT)
import pandas as pd
import os
import numpy
import h5py

LOG0_HEADER = ["index", "x", "y", "a", "b", "c", "d", "e", "f"]
MAX_SHIFT = 30
OVERLAP_THRESHOLD = 0.9
DEBUG_SDK_PATH="/data/informatics/CAM/roi_filter/allensdk/"
DEBUG_SCRIPT=os.path.join(DEBUG_SDK_PATH, "allensdk", "internal",
                          "pipeline_modules", "run_roi_filter.py")
DEBUG_OUTPUT_DIRECTORY="/data/informatics/CAM/roi_filter/"


def get_motion_filepath(experiment_id):
    return lu.query("""
select CONCAT(wkf.storage_directory, wkf.filename) as path
from well_known_files wkf
join well_known_file_types wkft on wkft.id = wkf.well_known_file_type_id
join ophys_experiments oe on oe.id = wkf.attachable_id
where oe.id = {} and
wkft.name like 'OphysMotionXyOffsetData'""".format(experiment_id))[0]["path"]


def get_segmentation_filepath(experiment_id, file_type):
    return lu.query("""
select CONCAT(wkf.storage_directory, wkf.filename) as path
from well_known_files wkf
join well_known_file_types wkft on wkft.id = wkf.well_known_file_type_id
join ophys_cell_segmentation_runs ocsr on ocsr.id = wkf.attachable_id
join ophys_experiments oe on oe.id = ocsr.ophys_experiment_id
where oe.id = {} and wkft.name like '{}'
and ocsr.current = 't'""".format(experiment_id, file_type))[0]["path"]


def get_model_info(experiment_id):
    res = lu.query("""
select CONCAT(wkf.storage_directory, wkf.filename) as path, wkf.id
from ophys_experiments oe
join ophys_sessions os on os.id = oe.ophys_session_id
join projects p on p.id = os.project_id
join well_known_files wkf on wkf.attachable_id = p.id
join well_known_file_types wkft on wkft.id = wkf.well_known_file_type_id
where oe.id = {} and
wkft.name = 'RoiLabelModel'""".format(experiment_id))[0]
    return res["path"], res["id"]


def get_genotype_info(experiment_id, code):
    res = lu.query("""
select g.name
from ophys_experiments oe
join ophys_sessions os on os.id = oe.ophys_session_id
join specimens s on s.id = os.specimen_id
join donors d on d.id = s.donor_id
join donors_genotypes dg on dg.donor_id = d.id
join genotypes g on g.id = dg.genotype_id
join genotype_types gt on gt.id = g.genotype_type_id
where oe.id = {} and gt.code like '{}'""".format(experiment_id, code))
    output = set()
    for line in res:
        output.add(line["name"])
    return list(output)


def create_input_data(experiment_id):
    data = {}
    data["log_0"] = get_motion_filepath(experiment_id)
    model, model_id = get_model_info(experiment_id)
    data["roi_label_model"] = model
    data["roi_label_model_id"] = model_id
    data["targeted_structure_id"] = lu.query("""
select targeted_structure_id
from ophys_experiments oe
where oe.id = {}""".format(experiment_id))[0]["targeted_structure_id"]
    data["imaging_depth"] = lu.query("""
select calculated_depth
from ophys_experiments oe
where oe.id = {}""".format(experiment_id))[0]["calculated_depth"]
    data["drivers"] = get_genotype_info(experiment_id, "D")
    data["reporters"] = get_genotype_info(experiment_id, "R")
    data["max_int_file"] = get_segmentation_filepath(
        experiment_id, "OphysSegmentationMaskData")
    data["object_list"] = get_segmentation_filepath(
        experiment_id, "OphysSegmentationObjects")
    return data


def debug(experiment_id, local=False, sdk_path=DEBUG_SDK_PATH,
          script=DEBUG_SCRIPT, output_directory=DEBUG_OUTPUT_DIRECTORY):
    input_data = create_input_data(experiment_id)
    exp_dir = os.path.join(output_directory, str(experiment_id))
    run_module(script,
               input_data,
               exp_dir,
               sdk_path=sdk_path,
               local=local)


def load_object_list(filename):
    '''Load the object list file.'''
    dataframe = pd.read_csv(filename)
    dataframe.columns = [column.strip() for column in dataframe.columns]
    return dataframe


def load_log0(filename):
    '''Load the log0 file.'''
    return pd.read_csv(filename, header=None, names=LOG0_HEADER)


def load_all_input(data):
    '''Load all input data from the input json.'''
    try:
        object_list_file = data["object_list"]
        object_data = load_object_list(object_list_file)
    except KeyError:
        logging.error("Input json missing object_list")
        raise
    except IOError:
        logging.error("Could not read object list file %s", object_list_file)
        raise

    try:
        log0_file = data["log_0"]
        motion_data = load_log0(log0_file)
    except KeyError:
        logging.error("Input json missing log_0")
        raise
    except IOError:
        logging.error("Could not read log_0 file %s", log0_file)
        raise

    try:
        maxint_file = data["max_int_file"]
        with h5py.File(maxint_file, "r") as f:
            segmentation_stack = f["data"][...]
    except KeyError:
        logging.error("Input json missing max_int_file")
        raise
    except IOError:
        logging.error("Could not read max_int_file file %s", maxint_file)
        raise

    try:
        model_file = data["roi_label_model"]
        classifier = roi_filter.ROIClassifier.from_file(model_file)
    except KeyError:
        logging.error("Input json missing roi_label_model")
        raise
    except IOError:
        logging.error("Could not read roi_label_model file %s", model_file)
        raise

    try:
        depth = float(data["imaging_depth"])
    except KeyError:
        logging.error("Input json missing imaging_depth")
        raise
    except ValueError:
        logging.error("Invalid depth %s", data["imaging_depth"])
        raise

    try:
        structure_id = str(data["targeted_structure_id"])
    except KeyError:
        logging.error("Input json missing targeted_structure_id")
        raise

    try:
        model_id = data["roi_label_model_id"]
    except KeyError:
        logging.error("Input json missing roi_label_model_id")
        raise

    try:
        drivers = data["drivers"]
    except KeyError:
        logging.error("Input json drivers")
        raise

    try:
        reporters = data["reporters"]
    except KeyError:
        logging.error("Input json missing reporters")
        raise

    border = roi_filter_utils.calculate_max_border(motion_data, MAX_SHIFT)
    rois = roi_filter_utils.get_rois(segmentation_stack, border)
    rois = roi_filter_utils.order_rois_by_object_list(object_data, rois)

    result = {"model_id": model_id,
              "classifier": classifier,
              "object_data": object_data,
              "depth": depth,
              "structure_id": structure_id,
              "drivers": drivers,
              "reporters": reporters,
              "border": border,
              "rois": rois}
    return result


def create_output_data(rois, model_id, border, excluded,
                       unexpected_features):
    data = {}
    data["motion_border"] = {"x0": border[RIGHT_SHIFT],
                             "y0": border[DOWN_SHIFT],
                             "x1": border[LEFT_SHIFT],
                             "y1": border[UP_SHIFT]}
    data["roi_label_model_id"] = model_id
    data["unexpected_features"] = unexpected_features
    if rois:
        data["image"] = {"width": rois[0].img_cols,
                         "height": rois[0].img_rows}
    json_rois = {}
    for i, roi in enumerate(rois):
        json_roi = {}
        json_roi["x"] = roi.x
        json_roi["y"] = roi.y
        json_roi["width"] = roi.width
        json_roi["height"] = roi.height
        json_roi["mask"] = roi.mask
        json_roi["mask_page"] = roi.mask_group
        json_roi["exclusion_labels"] = roi.labels
        if roi.labels:
            json_roi["valid"] = False
        else:
            json_roi["valid"] = True
        # for backwards compatibility
        json_roi["exclude_code"] = excluded[i]
        json_rois[roi.label] = json_roi
    data["rois"] = json_rois
    return data


def main():
    mod = PipelineModule("Filter Ophys ROIs produced from cell segmentation.")

    input_data = mod.input_data()
    data = load_all_input(input_data)
    model_id = data["model_id"]
    classifier = data["classifier"]
    object_data = data["object_data"]
    depth = data["depth"]
    structure_id = data["structure_id"]
    drivers = data["drivers"]
    reporters = data["reporters"]
    border = data["border"]
    rois = data["rois"]

    label_array = classifier.get_labels(object_data, depth, structure_id,
                                        drivers, reporters)
    rois = roi_filter.apply_labels(rois, label_array, classifier.label_names)

    rois = roi_filter.label_unions_and_duplicates(rois, OVERLAP_THRESHOLD)

    output_data = create_output_data(rois, model_id, border,
                                     object_data["eXcluded"],
                                     classifier.unexpected_features)

    mod.write_output_data(output_data)

if __name__ == "__main__": main()



import argparse
import os
import sys
import subprocess
import logging
import numpy as np
import allensdk.core.json_utilities as json_utilities
from fit_stage_1 import SEEDS, FIT_BASE_DIR, MPIEXEC
import allensdk.internal.model.biophysical.optimize as optimize

FIT_TYPES = {"f6": "f9", "f12": "f13"}
DEFAULT_NUM_PROCESSES = 240

_fit_stage_2_log = logging.getLogger('allensdk.model.biophysical.fit_stage_2')

def prepare_stage_2(output_directory):
    config_base_data = json_utilities.read(os.path.join(FIT_BASE_DIR, 'config_base.json'))

    jobs = []

    for fit_type in FIT_TYPES:
        best_error = 1e12
        best_seed = 0
        
        fit_type_dir = os.path.join(output_directory, fit_type)

        if not os.path.exists(fit_type_dir):
            _fit_stage_2_log.debug("fit type directory does not exist for cell: %s" % fit_type_dir)
            continue

        for seed in SEEDS:
            hof_fit_file = os.path.join(fit_type_dir, "s%d" % seed, "final_hof_fit.txt")
            if not os.path.exists(hof_fit_file):
                _fit_stage_2_log.debug("hof fit file does not exist for seed: %d" % (seed))
                continue

            hof_fit = np.loadtxt(hof_fit_file)
            best_for_seed = np.min(hof_fit)
            if best_for_seed < best_error:
                best_seed = seed
                best_error = best_for_seed

        _fit_stage_2_log.debug("Best error for fit type %s is %f for seed %d" % (fit_type, best_error, best_seed))

        start_pop_file = os.path.join(fit_type_dir, "s%d" % best_seed, "final_hof.txt")
        new_fit_type_dir = os.path.join(output_directory, FIT_TYPES[fit_type])

        for seed in SEEDS:
            seed_dir = os.path.join(new_fit_type_dir, "s%d" % seed)
            if not os.path.exists(seed_dir):
                os.makedirs(seed_dir)

        target_file = os.path.join(output_directory, "target.json")
        target_data = json_utilities.read(target_file)
        has_apic = "apic" in target_data["passive"][0]["cm"]

        config = config_base_data.copy()
        config_path = os.path.join(new_fit_type_dir, "config.json")
        config["biophys"][0]["model_file"] = [ target_file, config_path]

        if has_apic:
            fit_style_file = os.path.join(FIT_BASE_DIR, "fit_styles", FIT_TYPES[fit_type] + "_fit_style.json")
        else:
            fit_style_file = os.path.join(FIT_BASE_DIR, "fit_styles", FIT_TYPES[fit_type] + "_noapic_fit_style.json")
        config["biophys"][0]["model_file"].append( fit_style_file )

        config["manifest"].append({"type": "dir", "spec": new_fit_type_dir, "key": "FITDIR"})
        config["manifest"].append({"type": "file", "spec": start_pop_file, "key": "STARTPOP"})

        json_utilities.write(config_path, config)

        for seed in SEEDS:
            logfile = os.path.join(new_fit_type_dir, 's%d' % seed, 'stage_2.log')

            jobs.append({
                    'config_path': os.path.abspath(config_path),
                    'fit_type': fit_type,
                    'log': os.path.abspath(logfile),
                    'seed': seed,
                    'num_processes': DEFAULT_NUM_PROCESSES
                    })

    return jobs


def run_stage_2(jobs):
    for job in jobs:
        args = [MPIEXEC,
                '-np', str(job['num_processes']),
                sys.executable,
                '-m',
                optimize.__name__,
                str(job['seed']),
                job['config_path'],
                str(optimize.DEFAULT_NGEN),
                str(optimize.DEFAULT_MU)]
        _fit_stage_2_log.debug(args)
        with open(job['log'], "w") as outfile:
            subprocess.call(args, stdout=outfile)


def main():
    parser = argparse.ArgumentParser(description='Set up DEAP-style fit for second stage')
    parser.add_argument('--output_dir', required=True)
    parser.add_argument('specimen_id', type=int)
    args = parser.parse_args()

    output_directory = os.path.join(args.output_dir, 'specimen_%d' % args.specimen_id)

    jobs = prepare_stage_2(output_directory)
    run_stage_2(jobs)

if __name__ == "__main__": 
    main()

